"""Kernels for SPD matrices."""

import numpy as np

from .base import invsqrtm, logm
from .mean import mean_riemann
from .utils import check_function


def kernel_euclid(X, Y=None, *, reg=1e-10, **kwargs):
    r"""Euclidean kernel between two sets of matrices.

    Calculates the Euclidean kernel matrix :math:`\mathbf{K}` of inner products
    of two sets :math:`\mathbf{X}` and :math:`\mathbf{Y}` of matrices in
    :math:`\mathbb{R}^{n \times m}` by calculating pairwise products:

    .. math::
        \mathbf{K}_{i,j} = \text{tr}(\mathbf{X}_i^T \mathbf{Y}_j)

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, m)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, m), default=None
        Second set of matrices. If None, Y is set to X.
    reg : float, default=1e-10
        Regularization parameter to mitigate numerical errors in kernel
        matrix estimation.

    Returns
    -------
    K : ndarray, shape (n_matrices_X, n_matrices_Y)
        The Euclidean kernel matrix between X and Y.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    kernel
    """
    def kernelfct(X, Cref):
        return X

    return _apply_matrix_kernel(kernelfct, X, Y, reg=reg)


def kernel_logeuclid(X, Y=None, *, reg=1e-10, **kwargs):
    r"""Log-Euclidean kernel between two sets of SPD matrices.

    Calculates the Log-Euclidean kernel matrix :math:`\mathbf{K}` of inner
    products of two sets :math:`\mathbf{X}` and :math:`\mathbf{Y}` of SPD
    matrices in :math:`\mathbb{R}^{n \times n}` by calculating pairwise
    products [1]_:

    .. math::
        \mathbf{K}_{i,j} = \text{tr}(\log(\mathbf{X}_i) \log(\mathbf{Y}_j))

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of SPD matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of SPD matrices. If None, Y is set to X.
    reg : float, default=1e-10
        Regularization parameter to mitigate numerical errors in kernel
        matrix estimation.

    Returns
    -------
    K : ndarray, shape (n_matrices_X, n_matrices_Y)
        The Log-Euclidean kernel matrix between X and Y.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    kernel

    References
    ----------
    .. [1] `Classification of covariance matrices using a Riemannian-based
        kernel for BCI applications
        <https://hal.archives-ouvertes.fr/hal-00820475/>`_
        A. Barachant, S. Bonnet, M. Congedo and C. Jutten. Neurocomputing,
        Elsevier, 2013, 112, pp.172-178.
    """
    def kernelfct(X, Cref):
        return logm(X)

    return _apply_matrix_kernel(kernelfct, X, Y, reg=reg)


def kernel_riemann(X, Y=None, *, Cref=None, reg=1e-10):
    r"""Affine-invariant Riemannian kernel between two sets of SPD matrices.

    Calculates the affine-invariant Riemannian kernel matrix :math:`\mathbf{K}`
    of inner products of two sets :math:`\mathbf{X}` and :math:`\mathbf{Y}` of
    SPD matrices in :math:`\mathbb{R}^{n \times n}` on tangent space at
    :math:`\mathbf{C}_\text{ref}` by calculating pairwise products [1]_:

    .. math::
        \mathbf{K}_{i,j} = \text{tr}( \log( \mathbf{C}_\text{ref}^{-1/2}
        \mathbf{X}_i \mathbf{C}_\text{ref}^{-1/2} )
        \log( \mathbf{C}_\text{ref}^{-1/2} \mathbf{Y}_j
        \mathbf{C}_\text{ref}^{-1/2}) )

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of SPD matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of SPD matrices. If None, Y is set to X.
    Cref : None | ndarray, shape (n, n), default=None
        Reference point for the tangent space and inner product calculation.
        If None, Cref is calculated as the Riemannian mean of X.
    reg : float, default=1e-10
        Regularization parameter to mitigate numerical errors in kernel
        matrix estimation.

    Returns
    -------
    K : ndarray, shape (n_matrices_X, n_matrices_Y)
        The affine-invariant Riemannian kernel matrix between X and Y.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    kernel

    References
    ----------
    .. [1] `Classification of covariance matrices using a Riemannian-based
        kernel for BCI applications
        <https://hal.archives-ouvertes.fr/hal-00820475/>`_
        A. Barachant, S. Bonnet, M. Congedo and C. Jutten. Neurocomputing,
        Elsevier, 2013, 112, pp.172-178.
    """
    def kernelfct(X, Cref):
        if Cref is None:
            Cref = mean_riemann(X)

        C_invsq = invsqrtm(Cref)
        X_ = logm(C_invsq @ X @ C_invsq)
        return X_

    return _apply_matrix_kernel(kernelfct, X, Y, Cref=Cref, reg=reg)


###############################################################################


def _check_dimensions(X, Y, Cref):
    """Check for matching dimensions in X, Y and Cref."""
    if not isinstance(Y, type(None)):
        assert Y.shape[1:] == X.shape[1:], f"Dimension of matrices in Y must "\
                                           f"match dimension of matrices in " \
                                           f"X. Expected {X.shape[1:]}, got " \
                                           f"{Y.shape[1:]}."

    if not isinstance(Cref, type(None)):
        assert Cref.shape == X.shape[1:], f"Dimension of Cref must match " \
                                          f"dimension of matrices in X. " \
                                          f"Expected {X.shape[1:]}, got " \
                                          f"{Cref.shape}."


def _apply_matrix_kernel(kernel_fct, X, Y=None, *, Cref=None, reg=1e-10):
    """Apply a matrix kernel function."""
    _check_dimensions(X, Y, Cref)
    n_matrices_X, n, n = X.shape

    X_ = kernel_fct(X, Cref)

    if isinstance(Y, type(None)) or np.array_equal(X, Y):
        Y_ = X_
    else:
        Y_ = kernel_fct(Y, Cref)

    # calculate scalar products: K[i,j] = np.trace(X_[i]^T @ Y_[j])
    X_T = X_.transpose((0, 2, 1))
    K = np.einsum('acb,dbc->ad', X_T, Y_, optimize=True)

    # regularization due to numerical errors
    if np.array_equal(X_, Y_):
        K.flat[:: n_matrices_X + 1] += reg

    return K


kernel_functions = {
    "euclid": kernel_euclid,
    "logeuclid": kernel_logeuclid,
    "riemann": kernel_riemann,
}


def kernel(X, Y=None, *, Cref=None, metric="riemann", reg=1e-10):
    """Kernel matrix between matrices according to a specified metric.

    Calculates the kernel matrix K of inner products of two sets X and Y of
    matrices on the tangent space at Cref according to a specified metric.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    Cref : None | ndarray, shape (n, n), default=None
        Reference point for the tangent space and inner product
        calculation. Only used if metric="riemann".
    metric : string | callable, default="riemann"
        Metric used for tangent space and mean estimation, can be:
        "euclid", "logeuclid", "riemann", or a callable function.
    reg : float, default=1e-10
        Regularization parameter to mitigate numerical errors in kernel
        matrix estimation, to provide a positive-definite kernel matrix.

    Returns
    -------
    K : ndarray, shape (n_matrices_X, n_matrices_Y)
        The kernel matrix between X and Y.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    kernel_euclid
    kernel_logeuclid
    kernel_riemann
    """
    kernel_function = check_function(metric, kernel_functions)
    K = kernel_function(X, Y, Cref=Cref, reg=reg)
    return K



from ._version import __version__
from . import classification
from . import tangentspace
from . import channelselection
from . import estimation
from . import spatialfilters
from . import clustering
from . import stats
from . import embedding
from . import preprocessing
from . import datasets


__all__ = [
    '__version__',
    'classification',
    'tangentspace',
    'channelselection',
    'estimation',
    'spatialfilters',
    'clustering',
    'stats',
    'embedding',
    'preprocessing',
    'datasets',
]



import numpy as np
import pytest
from pytest import approx

from conftest import get_metrics
from pyriemann.tangentspace import TangentSpace, FGDA


@pytest.mark.parametrize("tspace", [TangentSpace, FGDA])
class TangentSpaceTestCase:
    def test_tangentspace(self, tspace, get_mats, get_labels):
        n_classes, n_matrices, n_channels = 2, 6, 3
        covmats = get_mats(n_matrices, n_channels, "spd")
        labels = get_labels(n_matrices, n_classes)
        self.clf_transform(tspace, covmats, labels)
        self.clf_fit_transform(tspace, covmats, labels)
        self.clf_fit_transform_independence(tspace, covmats, labels)
        if tspace is TangentSpace:
            self.clf_transform_wo_fit(tspace, covmats)
            self.clf_inversetransform(tspace, covmats)


class TestTangentSpace(TangentSpaceTestCase):
    def clf_transform(self, tspace, covmats, labels):
        n_matrices, n_channels, n_channels = covmats.shape
        ts = tspace().fit(covmats, labels)
        Xtr = ts.transform(covmats)
        if tspace is TangentSpace:
            n_ts = (n_channels * (n_channels + 1)) // 2
            assert Xtr.shape == (n_matrices, n_ts)
        else:
            assert Xtr.shape == (n_matrices, n_channels, n_channels)

    def clf_fit_transform(self, tspace, covmats, labels):
        n_matrices, n_channels, n_channels = covmats.shape
        ts = tspace()
        Xtr = ts.fit_transform(covmats, labels)
        if tspace is TangentSpace:
            n_ts = (n_channels * (n_channels + 1)) // 2
            assert Xtr.shape == (n_matrices, n_ts)
        else:
            assert Xtr.shape == (n_matrices, n_channels, n_channels)

    def clf_fit_transform_independence(self, tspace, covmats, labels):
        n_matrices, n_channels, n_channels = covmats.shape
        ts = tspace()
        ts.fit(covmats, labels)
        # retraining with different size should erase previous fit
        new_covmats = covmats[:, :-1, :-1]
        ts.fit(new_covmats, labels)
        # fit_transform should work as well
        ts.fit_transform(covmats, labels)

    def clf_transform_wo_fit(self, tspace, covmats):
        n_matrices, n_channels, n_channels = covmats.shape
        n_ts = (n_channels * (n_channels + 1)) // 2
        ts = tspace()
        Xtr = ts.transform(covmats)
        assert Xtr.shape == (n_matrices, n_ts)

    def clf_inversetransform(self, tspace, covmats):
        n_matrices, n_channels, n_channels = covmats.shape
        ts = tspace().fit(covmats)
        Xtr = ts.transform(covmats)
        covinv = ts.inverse_transform(Xtr)
        assert covinv == approx(covmats)


@pytest.mark.parametrize("fit", [True, False])
@pytest.mark.parametrize("tsupdate", [True, False])
@pytest.mark.parametrize("metric_mean", get_metrics())
@pytest.mark.parametrize("metric_map", ["euclid", "logeuclid", "riemann"])
def test_TangentSpace_init(fit, tsupdate, metric_mean, metric_map, get_mats):
    n_matrices, n_channels = 4, 3
    n_ts = (n_channels * (n_channels + 1)) // 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    ts = TangentSpace(
        metric={"mean": metric_mean, "map": metric_map},
        tsupdate=tsupdate
    )
    if fit:
        ts.fit(covmats)
    Xtr = ts.transform(covmats)
    assert Xtr.shape == (n_matrices, n_ts)


@pytest.mark.parametrize("tsupdate", [True, False])
@pytest.mark.parametrize("metric_mean", get_metrics())
@pytest.mark.parametrize("metric_map", ["euclid", "logeuclid", "riemann"])
def test_FGDA_init(tsupdate, metric_mean, metric_map, get_mats, get_labels):
    n_classes, n_matrices, n_channels = 2, 6, 3
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    ts = FGDA(
        metric={"mean": metric_mean, "map": metric_map},
        tsupdate=tsupdate
    )
    ts.fit(covmats, labels)
    Xtr = ts.transform(covmats)
    assert Xtr.shape == (n_matrices, n_channels, n_channels)


@pytest.mark.parametrize("tspace", [TangentSpace, FGDA])
@pytest.mark.parametrize("metric", [42, "faulty", {"foo": "bar"}])
def test_metric_wrong_keys(tspace, metric, get_mats, get_labels):
    with pytest.raises((TypeError, KeyError, ValueError)):
        n_classes, n_matrices, n_channels = 2, 6, 3
        labels = get_labels(n_matrices, n_classes)
        covmats = get_mats(n_matrices, n_channels, "spd")
        clf = tspace(metric=metric)
        clf.fit(covmats, labels).transform(covmats)


def test_TS_vecdim_error():
    n_matrices, n_ts = 4, 6
    ts = TangentSpace()
    with pytest.raises(ValueError):
        tsvectors_wrong = np.empty((n_matrices, n_ts + 1))
        ts.transform(tsvectors_wrong)


def test_TS_matdim_error():
    n_matrices, n_channels = 4, 3
    ts = TangentSpace()
    with pytest.raises(ValueError):
        not_square_mat = np.empty((n_matrices, n_channels, n_channels + 1))
        ts.transform(not_square_mat)
    with pytest.raises(ValueError):
        too_many_dim = np.empty((1, 2, 3, 4))
        ts.transform(too_many_dim)



import numpy as np
from numpy.testing import assert_array_equal
import pytest
from pytest import approx

from pyriemann.utils.ajd import ajd, rjd, ajd_pham, uwedge


@pytest.mark.parametrize(
    "method, algo",
    [
        ("rjd", rjd),
        ("ajd_pham", ajd_pham),
        ("uwedge", uwedge),
        (uwedge, uwedge),
    ]
)
def test_ajd(method, algo, get_mats):
    """Test ajd algos"""
    eps, n_iter_max = 1e-6, 100
    n_matrices, n_channels = 10, 3
    mats = get_mats(n_matrices, n_channels, "spd")

    V, D = ajd(mats, method=method, eps=eps, n_iter_max=n_iter_max)
    assert V.shape == (n_channels, n_channels)
    assert D.shape == (n_matrices, n_channels, n_channels)

    if method == "rjd":
        assert D == approx(V.T @ mats @ V)
        assert V.T @ V == approx(np.eye(n_channels))  # check orthogonality
    else:
        assert D == approx(V @ mats @ V.T)

    V_, D_ = algo(mats, eps=eps, n_iter_max=n_iter_max)
    assert np.all(V == V_)
    assert np.all(D == D_)


@pytest.mark.parametrize("method", ["rjd", "ajd_pham", "uwedge"])
@pytest.mark.parametrize("init", [True, False])
def test_ajd_init(method, init, get_mats_params):
    """Test init for ajd algos"""
    n_matrices, n_channels = 9, 4
    mats, _, evecs = get_mats_params(n_matrices, n_channels, "spd")
    if init:
        ajd(mats, method=method, init=evecs)
    else:
        ajd(mats, method=method)


@pytest.mark.parametrize("method", ["rjd", "ajd_pham", "uwedge"])
def test_ajd_init_error(method, get_mats):
    """Test init errors for ajd algos"""
    n_matrices, n_channels = 4, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):  # not 2D array
        ajd(mats, method=method, init=np.ones((3, 2, 2)))
    with pytest.raises(ValueError):  # not square array
        ajd(mats, method=method, init=np.ones((3, 2)))
    with pytest.raises(ValueError):  # shape not equal to n_channels
        ajd(mats, method=method, init=np.ones((2, 2)))


@pytest.mark.parametrize("method", ["abc", 42])
def test_ajd_method_error(method):
    with pytest.raises(ValueError):
        ajd(np.ones((3, 2, 2)), method=method)


def test_ajd_pham(get_mats):
    """Test pham's ajd"""
    n_matrices, n_channels = 7, 4
    mats = get_mats(n_matrices, n_channels, "spd")
    V, D = ajd_pham(mats)
    assert D == approx(V @ mats @ V.conj().T)


def test_ajd_pham_weight_none_equivalent_uniform(get_mats):
    """Test pham's ajd weights: none is equivalent to uniform values"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    V, D = ajd_pham(mats)
    Vw, Dw = ajd_pham(mats, sample_weight=np.ones(n_matrices))
    assert_array_equal(V, Vw)  # same result as ajd_pham without weight
    assert_array_equal(D, Dw)

    ajd(mats, method="ajd_pham", sample_weight=np.ones(n_matrices))


def test_ajd_pham_weight_positive(get_mats):
    """Test pham's ajd weights: must be strictly positive"""
    n_matrices, n_channels = 4, 2
    mats = get_mats(n_matrices, n_channels, "spd")
    w = 1.23 * np.ones(n_matrices)
    with pytest.raises(ValueError):  # not strictly positive weight
        w[0] = 0
        ajd_pham(mats, sample_weight=w)


def test_ajd_pham_weight_zero(get_mats):
    """Setting one weight to almost 0 it's almost like not passing the mat"""
    n_matrices, n_channels = 5, 4
    mats = get_mats(n_matrices, n_channels, "spd")
    w = 4.32 * np.ones(n_matrices)
    V, D = ajd_pham(mats[1:], sample_weight=w[1:])
    w[0] = 1e-12
    Vw, Dw = ajd_pham(mats, sample_weight=w)
    assert V == approx(Vw, rel=1e-4, abs=1e-8)
    assert D == approx(Dw[1:], rel=1e-4, abs=1e-8)



import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest
from pytest import approx
from scipy.spatial.distance import euclidean, mahalanobis

from conftest import get_distances
from pyriemann.utils.distance import (
    distance_euclid,
    distance_harmonic,
    distance_kullback,
    distance_kullback_right,
    distance_kullback_sym,
    distance_logdet,
    distance_logeuclid,
    distance_riemann,
    distance_wasserstein,
    distance,
    pairwise_distance,
    distance_mahalanobis,
)
from pyriemann.utils.base import logm
from pyriemann.utils.geodesic import geodesic
from pyriemann.utils.test import is_sym


def get_dist_func():
    dist_func = [
        distance_euclid,
        distance_harmonic,
        distance_kullback,
        distance_kullback_right,
        distance_kullback_sym,
        distance_logdet,
        distance_logeuclid,
        distance_riemann,
        distance_wasserstein,
    ]
    for df in dist_func:
        yield df


def callable_sp_euclidean(A, B, squared=False):
    return euclidean(A.flatten(), B.flatten())


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "metric, dist",
    [
        ("euclid", distance_euclid),
        ("harmonic", distance_harmonic),
        ("kullback", distance_kullback),
        ("kullback_right", distance_kullback_right),
        ("kullback_sym", distance_kullback_sym),
        ("logdet", distance_logdet),
        ("logeuclid", distance_logeuclid),
        ("riemann", distance_riemann),
        ("wasserstein", distance_wasserstein),
        (callable_sp_euclidean, distance_euclid),
    ],
)
def test_distances_metric(kind, metric, dist, get_mats):
    """Test distance for metric"""
    n_matrices, n_channels = 2, 3
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]
    d = distance(A, B, metric=metric)
    assert d == approx(dist(A, B))
    assert np.isreal(d)


def test_distances_metric_error(get_mats):
    n_matrices, n_channels = 2, 2
    A = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        distance(A[0], A[1], metric="universe")
    with pytest.raises(ValueError):
        distance(A[0], A[1], metric=42)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_squared(kind, dist, get_mats):
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]
    assert dist(A, B, squared=True) == approx(dist(A, B) ** 2)


@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_all_error(dist, get_mats):
    n_matrices, n_channels = 3, 3
    A = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        dist(A, A[0])


@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_all_ndarray(dist, get_mats):
    n_matrices, n_channels = 5, 3
    A = get_mats(n_matrices, n_channels, "spd")
    B = get_mats(n_matrices, n_channels, "spd")
    assert isinstance(dist(A[0], B[0]), float)  # 2D arrays
    assert dist(A, B).shape == (n_matrices,)  # 3D arrays

    n_sets = 5
    C = np.asarray([A for _ in range(n_sets)])
    D = np.asarray([B for _ in range(n_sets)])
    assert dist(C, D).shape == (n_sets, n_matrices,)  # 4D arrays


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_all_geodesic(kind, dist, get_mats):
    n_matrices, n_channels = 2, 6
    mats = get_mats(n_matrices, n_channels, kind)
    A, C = mats[0], mats[1]
    B = geodesic(A, C, alpha=0.5)
    assert dist(A, B) < dist(A, C)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_all_separability(kind, dist, get_mats):
    n_matrices, n_channels = 1, 6
    mats = get_mats(n_matrices, n_channels, kind)
    assert dist(mats[0], mats[0]) == approx(0, abs=2e-7)
    assert dist(np.eye(n_channels), np.eye(n_channels)) == approx(0)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "dist", [
        distance_euclid,
        distance_harmonic,
        distance_kullback_sym,
        distance_logdet,
        distance_logeuclid,
        distance_riemann,
        distance_wasserstein,
    ]
)
def test_distances_all_symmetry(kind, dist, get_mats):
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]
    assert dist(A, B) == approx(dist(B, A))


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("dist", get_dist_func())
def test_distances_all_triangle_inequality(kind, dist, get_mats):
    n_matrices, n_channels = 3, 4
    mats = get_mats(n_matrices, n_channels, kind)
    A, B, C = mats[0], mats[1], mats[2]
    assert dist(A, B) <= dist(A, C) + dist(C, B)


@pytest.mark.parametrize("complex_valued", [True, False])
def test_distance_euclid(rndstate, complex_valued):
    """Test Euclidean distance for generic matrices"""
    n_matrices, n_dim0, n_dim1 = 2, 3, 4
    mats = rndstate.randn(n_matrices, n_dim0, n_dim1)
    if complex_valued:
        mats = mats + 1j * rndstate.randn(n_matrices, n_dim0, n_dim1)
    A, B = mats[0], mats[1]
    distance_euclid(A, B)


@pytest.mark.parametrize("kind", ["real", "comp"])
def test_distance_harmonic(kind, get_mats):
    """Test harmonic distance for invertible matrices"""
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]
    distance_harmonic(A, B)


def test_distance_kullback_implementation(get_mats):
    n_matrices, n_channels = 2, 6
    mats = get_mats(n_matrices, n_channels, "spd")
    A, B = mats[0], mats[1]
    d = 0.5*(np.trace(np.linalg.inv(B) @ A) - n_channels
             + np.log(np.linalg.det(B) / np.linalg.det(A)))
    assert distance_kullback(A, B) == approx(d)


def test_distance_logdet_implementation(get_mats):
    n_matrices, n_channels = 2, 6
    mats = get_mats(n_matrices, n_channels, "spd")
    A, B = mats[0], mats[1]
    d = np.sqrt(np.log(np.linalg.det((A + B) / 2.0))
                - 0.5 * np.log(np.linalg.det(A)*np.linalg.det(B)))
    assert distance_logdet(A, B) == approx(d)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_distance_riemann_properties(kind, get_mats):
    n_channels = 6
    mats = get_mats(2, n_channels, kind)
    A, B = mats[0], mats[1]
    dist_AB = distance_riemann(A, B)

    # exponential metric increasing property, Eq(6.8) in [Bhatia2007]
    assert dist_AB >= np.linalg.norm(logm(A) - logm(B))

    # invariance under inversion
    assert dist_AB == approx(
        distance_riemann(np.linalg.inv(A), np.linalg.inv(B))
    )

    # congruence-invariance
    W = np.random.normal(size=(n_channels, n_channels))  # must be invertible
    WAW, WBW = W @ A @ W.T, W @ B @ W.T
    assert dist_AB == approx(distance_riemann(WAW, WBW))

    # proportionality, Eq(6.12) in [Bhatia2007]
    alpha = np.random.uniform()
    dist_1 = distance_riemann(A, geodesic(A, B, alpha, metric='riemann'))
    dist_2 = alpha * distance_riemann(A, B)
    assert dist_1 == approx(dist_2)


@pytest.mark.parametrize("dist, dfunc", zip(get_distances(), get_dist_func()))
def test_distance_wrapper(dist, dfunc, get_mats):
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, "spd")
    A, B = mats[0], mats[1]
    assert distance(A, B, metric=dist) == dfunc(A, B)


@pytest.mark.parametrize("dist", get_dist_func())
def test_distance_wrapper_between_set_and_matrix(dist, get_mats):
    n_matrices, n_channels = 10, 4
    mats = get_mats(n_matrices, n_channels, "spd")
    assert distance(mats, mats[-1], metric=dist).shape == (n_matrices, 1)

    n_sets = 5
    covs_4d = np.asarray([mats for _ in range(n_sets)])
    with pytest.raises(ValueError):
        distance(covs_4d, mats, metric=dist)


@pytest.mark.parametrize("dist", get_distances())
@pytest.mark.parametrize("Y", [None, True])
@pytest.mark.parametrize("squared", [False, True])
def test_pairwise_distance_matrix(get_mats, dist, Y, squared):
    n_matrices_X, n_matrices_Y, n_channels = 6, 4, 5
    X = get_mats(n_matrices_X, n_channels, "spd")
    if Y is None:
        n_matrices_Y = n_matrices_X
        Y_ = X
    else:
        Y = get_mats(n_matrices_Y, n_channels, "spd")
        Y_ = Y

    pdist = pairwise_distance(X, Y, metric=dist, squared=squared)
    assert pdist.shape == (n_matrices_X, n_matrices_Y)

    for i in range(n_matrices_X):
        for j in range(n_matrices_Y):
            assert np.isclose(pdist[i, j],
                              distance(X[i],
                                       Y_[j],
                                       metric=dist,
                                       squared=squared),
                              atol=1e-5,
                              rtol=1e-5)

    if Y is None and dist not in ["kullback", "kullback_right"]:
        assert is_sym(pdist)
    else:
        assert not is_sym(pdist)


@pytest.mark.parametrize("complex_valued", [True, False])
def test_distance_mahalanobis(rndstate, complex_valued):
    n_channels, n_times = 3, 100
    X = rndstate.randn(n_channels, n_times)
    if complex_valued:
        X = X + 1j * rndstate.randn(n_channels, n_times)
    d = distance_mahalanobis(X, np.cov(X))
    assert d.shape == (n_times,)
    assert np.all(np.isreal(d))


@pytest.mark.parametrize("mean", [True, None])
def test_distance_mahalanobis_scipy(rndstate, get_mats, mean):
    """Test equivalence between pyriemann and scipy for real data"""
    n_channels, n_times = 3, 100
    X = rndstate.randn(n_channels, n_times)
    C = get_mats(1, n_channels, "spd")[0]

    Cinv = np.linalg.inv(C)
    y = np.zeros(n_channels)
    dist_sp = [mahalanobis(x, y, Cinv) for x in X.T]

    if mean:
        mean = np.zeros((n_channels, 1))
    else:
        mean = None
    dist_pr = distance_mahalanobis(X, C, mean=mean)

    assert_array_almost_equal(dist_sp, dist_pr)



import warnings
import numpy as np
from ..utils.distance import distance
from ..utils.utils import check_weights


def _project(X, U):
    """
    Project U on the tangent space of orthogonal matrices with base point X.
    """

    return X @ ((X.T @ U) - (X.T @ U).T) / 2


def _retract(X, v):
    """
    Retraction taking tangent vector v at base point X back to manifold of
    orthogonal matrices.
    """

    Q, R = np.linalg.qr(X + v)
    for i, ri in enumerate(np.diag(R)):
        if ri < 0:
            Q[:, i] = -1 * Q[:, i]
    return Q


def _loss(Q, X, Y, weights, metric='euclid'):
    """Loss function for estimating the rotation matrix in RPA."""

    return weights @ distance(X, Q @ Y @ Q.T, metric=metric)


def _grad(Q, X, Y, weights, metric='euclid'):
    """Gradient of loss function between class means."""

    if metric == 'euclid':
        return np.einsum('a,abc->bc', weights, -4 * (X - Q @ Y @ Q.T) @ Q @ Y)

    elif metric == 'riemann':
        M = np.linalg.inv(X) @ Q @ Y @ Q.T
        eigvals, eigvecs = np.linalg.eig(M)
        logeigvals = np.expand_dims(np.log(eigvals), -2)
        logM = (eigvecs * logeigvals) @ np.linalg.inv(eigvecs)
        return np.einsum('a,abc->bc', weights, 4 * logM @ Q)

    else:
        raise ValueError("RPA supports only 'euclid' and 'riemann' metrics.")


def _warm_start(X, Y, weights, metric='euclid'):
    """Smart initialization of the minimization procedure.

    The loss function being optimized is a weighted sum of loss functions with
    the same structure and for which we know the exact analytic solution [1]_.

    As such, a natural way to warm start the optimization procedure is to list
    the minimizers of "local" loss function and set as Q0 the matrix that
    yields the smallest value for the "global" loss function.

    Note that the initialization is the same for both 'euclid' and 'riemann'.

    References
    ----------
    .. [1] `Procrustes problems in Riemannian manifolds of positive definite
        matrices
        <https://hal.archives-ouvertes.fr/hal-02023293>`_
        R. Bhatia and M. Congedo, Linear Algebra and its Applications,
        Elsevier, 2019, 563, pp.440-445.
    """

    # obtain the solution of the local loss function
    def _get_local_solution(Xi, Yi):
        wX, qX = np.linalg.eig(Xi)
        idx = wX.argsort()[::-1]
        qX = qX[:, idx]
        wY, qY = np.linalg.eig(Yi)
        idx = wY.argsort()[::-1]
        qY = qY[:, idx]
        Qstar = qX @ qY.T
        return Qstar

    Q0_candidates, losses = [], []
    for i in range(len(X)):
        Q0_candidates.append(_get_local_solution(X[i], Y[i]))
        losses.append(_loss(Q0_candidates[i], X, Y, weights, metric=metric))

    i_min = np.argmin(losses)

    return Q0_candidates[i_min]


def _run_minimization(Q_ini, M_source, M_target, weights=None, metric='euclid',
                      tol_step=1e-9, maxiter=10_000, maxiter_linesearch=32):

    Q_1 = Q_ini

    # loop over iterations
    for _ in range(maxiter):

        # get the current value for the loss function
        F_1 = _loss(Q_1, M_target, M_source, weights, metric=metric)

        # get the direction of steepest descent
        direction = _project(
            Q_1, _grad(Q_1, M_target, M_source, weights, metric=metric)
        )

        # backtracking line search
        alpha = 1.0
        tau = 0.50
        r = 1e-4
        Q = _retract(Q_1, -alpha * direction)
        F = _loss(Q, M_target, M_source, weights, metric=metric)
        for _ in range(maxiter_linesearch):
            if F_1 - F > r * alpha * np.linalg.norm(direction)**2:
                break
            alpha = tau * alpha
            Q = _retract(Q_1, -alpha * direction)
            F = _loss(Q, M_target, M_source, weights, metric=metric)

        # test if the step size is small
        crit = np.linalg.norm(-alpha * direction)
        if crit <= tol_step:
            break

        # update variables for next iteration
        Q_1 = Q
        F_1 = F

    else:
        warnings.warn('Convergence not reached.')

    return Q, F


def _get_rotation_matrix(M_source, M_target, weights=None, metric='euclid',
                         tol_step=1e-9, maxiter=10_000, maxiter_linesearch=32):
    r"""Calculate rotation matrix for the Riemannian Procustes Analysis.

    Get the rotation matrix Q that minimizes the loss function:

    .. math::
        L(Q) = \sum_i w_i delta^2(M_{target_i}, Q M_{source_i} Q^T)

    The solution can then be used to transform the data points from the source
    domain so that their class means are close to the those from the target
    domain. This manifold optimization problem was first defined in Eq(35) of
    [1]_. The optimization procedure follows the usual setup for optimization
    on manifolds as described in [2]_.

    Parameters
    ----------
    M_source : ndarray, shape (n_classes, n, n)
        Set with the means of the n_classes from the source domain.
    M_target : ndarray, shape (n_classes, n, n)
        Set with the means of the n_classes from the target domain.
    weights : None | array, shape (n_classes,), default=None
        Weights for each class. If None, then give the same weight for each
        class.
    metric : {'euclid', 'riemann'}, default='euclid'
        Which type of distance to minimize between the class means, either
        'euclid' for Euclidean distance or 'riemann' for the affine-invariant
        Riemannian distance between SPD matrices.
    tol_step : float, default 1e-9
        Stopping criterion based on the norm of the descent direction.
    maxiter : int, default=10_000
        Maximum number of iterations in the optimization procedure.
    maxiter_linesearch : int, default=32
        Maximum number of iterations in the line search procedure.

    Returns
    -------
    Q : ndarray, shape (n, n)
        Orthogonal matrix that minimizes the distance between the class means
        for the source and target domains.

    References
    ----------
    .. [1] `Riemannian Procrustes analysis: transfer
        learning for brain-computer interfaces
        <https://hal.archives-ouvertes.fr/hal-01971856>`_
        PLC Rodrigues et al, IEEE Transactions on Biomedical Engineering,
        vol. 66, no. 8, pp. 2390-2401, December, 2018
    .. [2] `An introduction to optimization on smooth manifolds
        <https://www.nicolasboumal.net/book/>`_
        N. Boumal. To appear with Cambridge University Press. June, 2022

    Notes
    -----
    .. versionadded:: 0.4
    """

    if M_source.shape[0] != M_target.shape[0]:
        raise ValueError("The number of classes in each domain don't match")
    if M_source.shape[1:] != M_target.shape[1:]:
        raise ValueError("The number of channels in each domain don't match")

    weights = check_weights(weights, len(M_source))

    # initialize the solution with an educated guess
    Q_ini = _warm_start(M_target, M_source, weights, metric=metric)
    if np.linalg.det(Q_ini) < 0:  # make sure it is a rotation
        Q_ini[[0, 1], :] = Q_ini[[1, 0], :]

    # run the optimization procedure on rotation matrices
    Q_posdet, F_posdet = _run_minimization(
        Q_ini, M_source, M_target, weights, metric, tol_step, maxiter,
        maxiter_linesearch)

    # run the optimization procedure on reflection matrices
    Q_ini[[0, 1], :] = Q_ini[[1, 0], :]
    Q_negdet, F_negdet = _run_minimization(
        Q_ini, M_source, M_target, weights, metric, tol_step, maxiter,
        maxiter_linesearch)

    # check which of the options yield a smaller loss value
    if F_posdet < F_negdet:
        return Q_posdet
    else:
        return Q_negdet



import numpy as np
import pytest
from pytest import approx

from conftest import get_metrics
from pyriemann.utils.distance import distance_riemann
from pyriemann.utils.tangentspace import (
    exp_map_euclid, exp_map_logeuclid, exp_map_riemann,
    log_map_euclid, log_map_logeuclid, log_map_riemann,
    upper, unupper, tangent_space, untangent_space, transport
)


@pytest.mark.parametrize(
    "fun_map", [exp_map_euclid, exp_map_logeuclid, exp_map_riemann,
                log_map_euclid, log_map_logeuclid, log_map_riemann]
)
def test_maps_ndarray(fun_map, get_mats):
    """Test log and exp maps"""
    n_matrices, n_channels = 6, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    Xt = fun_map(mats, np.eye(n_channels))
    assert Xt.shape == (n_matrices, n_channels, n_channels)

    n_sets = 2
    covmats_4d = np.asarray([mats for _ in range(n_sets)])
    Xt = fun_map(covmats_4d, np.eye(n_channels))
    assert Xt.shape == (n_sets, n_matrices, n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "log_map, exp_map", zip(
        [log_map_euclid, log_map_logeuclid, log_map_riemann],
        [exp_map_euclid, exp_map_logeuclid, exp_map_riemann]
    )
)
def test_maps_log_exp(kind, log_map, exp_map, get_mats):
    """Test log then exp maps should be identity"""
    n_matrices, n_channels = 10, 3
    mats = get_mats(n_matrices, n_channels, kind)
    X, C = mats[:-1], mats[-1]
    assert exp_map(log_map(X, C), C) == approx(X)


@pytest.mark.parametrize("complex_valued", [True, False])
def test_map_euclid(rndstate, complex_valued):
    """Test Euclidean map for generic matrices"""
    n_matrices, n_dim0, n_dim1 = 5, 3, 4
    mats = rndstate.randn(n_matrices, n_dim0, n_dim1)
    if complex_valued:
        mats = mats + 1j * rndstate.randn(n_matrices, n_dim0, n_dim1)
    X, C = mats[:n_matrices - 1], mats[-1]
    assert exp_map_euclid(log_map_euclid(X, C), C) == approx(X)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_upper_and_unupper(kind, get_mats):
    """Test upper then unupper should be identity"""
    n_matrices, n_channels = 7, 3
    mats = get_mats(n_matrices, n_channels, kind)
    assert unupper(upper(mats)) == approx(mats)

    n_sets = 2
    mats_4d = np.asarray([mats for _ in range(n_sets)])
    assert unupper(upper(mats_4d)) == approx(mats_4d)


@pytest.mark.parametrize("metric", ["euclid", "logeuclid", "riemann"])
def test_tangent_space_ndarray(metric, get_mats):
    """Test tangent space projection"""
    n_matrices, n_channels = 6, 3
    n_ts = (n_channels * (n_channels + 1)) // 2
    X = get_mats(n_matrices, n_channels, "spd")
    Xts = tangent_space(X, np.eye(n_channels), metric=metric)
    assert Xts.shape == (n_matrices, n_ts)

    n_sets = 2
    X_4d = np.asarray([X for _ in range(n_sets)])
    Xts = tangent_space(X_4d, np.eye(n_channels), metric=metric)
    assert Xts.shape == (n_sets, n_matrices, n_ts)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_tangent_space_riemann_properties(kind, get_mats):
    n_matrices, n_channels = 2, 3
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]

    # equivalent definitions of Riemannian distance, Eq(7) in [Barachant2012]
    dist = distance_riemann(A, B)
    s = tangent_space(A, B, metric='riemann')
    assert dist == approx(np.linalg.norm(s))


@pytest.mark.parametrize("metric", ["euclid", "logeuclid", "riemann"])
def test_untangent_space_ndarray(metric, rndstate):
    """Test untangent space projection"""
    n_matrices, n_channels = 10, 3
    n_ts = (n_channels * (n_channels + 1)) // 2
    T = rndstate.randn(n_matrices, n_ts)
    X = untangent_space(T, np.eye(n_channels), metric=metric)
    assert X.shape == (n_matrices, n_channels, n_channels)

    n_sets = 2
    T_4d = np.asarray([T for _ in range(n_sets)])
    X = untangent_space(T_4d, np.eye(n_channels), metric=metric)
    assert X.shape == (n_sets, n_matrices, n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("metric", ["euclid", "logeuclid", "riemann"])
def test_tangent_and_untangent_space(kind, metric, get_mats):
    """Test tangent space projection then back-projection should be identity"""
    n_matrices, n_channels = 10, 3
    mats = get_mats(n_matrices, n_channels, kind)
    X, C = mats[:n_matrices - 1], mats[-1]
    X_t = tangent_space(X, C, metric=metric)
    X_ut = untangent_space(X_t, C, metric=metric)
    assert X_ut == approx(X)


@pytest.mark.parametrize("metric", get_metrics())
def test_transport(metric, get_mats):
    n_matrices, n_channels = 10, 3
    X = get_mats(n_matrices, n_channels, "spd")
    X_tr = transport(X, np.eye(n_channels), metric=metric)
    assert X_tr.shape == (n_matrices, n_channels, n_channels)



import numpy as np
import pytest

from pyriemann.datasets.sampling import (sample_gaussian_spd,
                                         generate_random_spd_matrix)
from pyriemann.utils.distance import distance_riemann
from pyriemann.utils.test import is_sym_pos_def as is_spd


@pytest.mark.parametrize("n_jobs", [1, -1])
@pytest.mark.parametrize("sampling_method", ['auto', 'slice', 'rejection'])
def test_sample_gaussian_spd_dim2(n_jobs, sampling_method):
    """Test Riemannian Gaussian sampling for dim=2."""
    n_matrices, n_dim, sigma = 5, 2, 1.
    mean = np.eye(n_dim)
    X = sample_gaussian_spd(n_matrices, mean, sigma, random_state=42,
                            n_jobs=n_jobs, sampling_method=sampling_method)
    assert X.shape == (n_matrices, n_dim, n_dim)  # X shape mismatch
    assert is_spd(X)  # X is an array of SPD matrices


@pytest.mark.parametrize("n_dim", [3, 4])
@pytest.mark.parametrize("n_jobs", [1, -1])
@pytest.mark.parametrize("sampling_method", ['auto', 'slice'])
def test_sample_gaussian_spd_dimsup(n_dim, n_jobs, sampling_method):
    """Test Riemannian Gaussian sampling for dim>2."""
    n_matrices, sigma = 5, 1.
    mean = np.eye(n_dim)
    X = sample_gaussian_spd(n_matrices, mean, sigma, random_state=42,
                            n_jobs=n_jobs, sampling_method=sampling_method)
    assert X.shape == (n_matrices, n_dim, n_dim)  # X shape mismatch
    assert is_spd(X)  # X is an array of SPD matrices


def test_sample_gaussian_spd_error():
    with pytest.raises(ValueError):  # unknown sampling method
        sample_gaussian_spd(5, np.eye(2), 1., sampling_method='blabla')
    with pytest.raises(ValueError):  # dim=3 not yet supported with rejection
        n_dim = 3
        sample_gaussian_spd(5, np.eye(n_dim), 1., sampling_method='rejection')


def test_generate_random_spd_matrix():
    """Test generating random SPD matrix"""
    n_dim = 4
    X = generate_random_spd_matrix(n_dim, random_state=None)
    assert X.shape == (n_dim, n_dim)  # X shape mismatch
    assert is_spd(X)  # X is a SPD matrix


@pytest.mark.parametrize("n_jobs", [1, -1])
def test_sigma_gaussian_spd(n_jobs):
    """Test sigma parameter from Riemannian Gaussian sampling."""
    n_matrices, n_dim, sig_1, sig_2 = 5, 4, 1., 2.
    mean = np.eye(n_dim)
    X1 = sample_gaussian_spd(
        n_matrices, mean, sig_1, random_state=42, n_jobs=n_jobs
        )
    X2 = sample_gaussian_spd(
        n_matrices, mean, sig_2, random_state=66, n_jobs=n_jobs
        )
    avg_d1 = np.mean([distance_riemann(X1_i, mean) for X1_i in X1])
    avg_d2 = np.mean([distance_riemann(X2_i, mean) for X2_i in X2])
    assert avg_d1 < avg_d2


def test_functions_error():
    n_matrices, n_dim = 3, 4
    mean, sigma = np.eye(n_dim), 2.
    with pytest.raises(ValueError):  # mean is not a matrix
        sample_gaussian_spd(n_matrices, np.ones(n_dim), sigma)
    with pytest.raises(ValueError):  # sigma is not a scalar
        sample_gaussian_spd(n_matrices, mean, np.ones(n_dim))
    with pytest.raises(ValueError):  # n_matrices is negative
        sample_gaussian_spd(-n_matrices, mean, sigma)
    with pytest.raises(ValueError):  # n_matrices is not an integer
        sample_gaussian_spd(4.2, mean, sigma)
    with pytest.raises(ValueError):  # n_dim is not an integer
        generate_random_spd_matrix(1.7)
    with pytest.raises(ValueError):  # n_dim is negative
        generate_random_spd_matrix(-n_dim)



"""Means of SPD/HPD matrices."""

from copy import deepcopy
import numpy as np
import warnings

from .ajd import ajd_pham
from .base import sqrtm, invsqrtm, logm, expm, powm
from .distance import distance_riemann
from .geodesic import geodesic_riemann
from .utils import check_weights, check_function


def _deprecate_covmats(covmats, X):
    if covmats is not None:
        print("DeprecationWarning: input covmats has been renamed into X and "
              "will be removed in 0.8.0.")
        X = covmats
    return X


def mean_ale(X=None, tol=10e-7, maxiter=50, sample_weight=None, covmats=None):
    """AJD-based log-Euclidean (ALE) mean of SPD matrices.

    Return the mean of a set of SPD matrices using the approximate joint
    diagonalization (AJD) based log-Euclidean (ALE) mean [1]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD matrices.
    tol : float, default=10e-7
        The tolerance to stop the gradient descent.
    maxiter : int, default=50
        The maximum number of iterations.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        ALE mean.

    Notes
    -----
    .. versionadded:: 0.2.4

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Approximate Joint Diagonalization and Geometric Mean of Symmetric
        Positive Definite Matrices
        <https://arxiv.org/abs/1505.07343>`_
        M. Congedo, B. Afsari, A. Barachant, M. Moakher. PLOS ONE, 2015
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, n, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)

    # init with AJD
    B = ajd_pham(X)[0]

    eye_n = np.eye(n)
    crit = np.inf
    for _ in range(maxiter):
        J = np.einsum('a,abc->bc', sample_weight, logm(B @ X @ B.conj().T))
        delta = np.real(np.diag(expm(J)))
        B = (np.abs(delta) ** -.5)[:, np.newaxis] * B

        crit = distance_riemann(eye_n, np.diag(delta))
        if crit <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    J = np.einsum('a,abc->bc', sample_weight, logm(B @ X @ B.conj().T))
    A = np.linalg.inv(B)
    M = A @ expm(J) @ A.conj().T
    return M


def mean_alm(X=None, tol=1e-14, maxiter=100, sample_weight=None, covmats=None):
    r"""Ando-Li-Mathias (ALM) mean of SPD/HPD matrices.

    Return the geometric mean recursively [1]_, generalizing from:

    .. math::
        \mathbf{M} = X_1^{\frac{1}{2}} (X_1^{-\frac{1}{2}}X_2^{\frac{1}{2}}
                     X_1^{-\frac{1}{2}})^{\frac{1}{2}} X_1^{\frac{1}{2}}

    and requiring a high number of iterations.
    Extremely slow, due to the recursive formulation.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    tol : float, default=10e-14
        The tolerance to stop the gradient descent.
    maxiter : int, default=100
        The maximum number of iterations.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        ALM mean.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Geometric Means
        <https://www.sciencedirect.com/science/article/pii/S0024379503008693>`_
        T. Ando, C.-K. Li, and R. Mathias. Linear Algebra and its Applications.
        Volume 385, July 2004, Pages 305-334.
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, _, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)

    if n_matrices == 2:
        alpha = sample_weight[1] / sample_weight[0] / 2
        M = geodesic_riemann(X[0], X[1], alpha=alpha)
        return M

    M = X
    M_iter = np.zeros_like(M)
    for _ in range(maxiter):
        for h in range(n_matrices):
            s = np.mod(np.arange(h, h + n_matrices - 1) + 1, n_matrices)
            M_iter[h] = mean_alm(M[s], sample_weight=sample_weight[s])

        norm_iter = np.linalg.norm(M_iter[0] - M[0], 2)
        norm_c = np.linalg.norm(M[0], 2)
        if (norm_iter / norm_c) < tol:
            break
        M = deepcopy(M_iter)
    else:
        warnings.warn("Convergence not reached")

    return M_iter.mean(axis=0)


def mean_euclid(X=None, sample_weight=None, covmats=None):
    r"""Mean of matrices according to the Euclidean metric.

    .. math::
        \mathbf{M} = \sum_i w_i \ \mathbf{X}_i

    This mean is also called arithmetic.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, m)
        Set of matrices.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, m)
        Euclidean mean.

    See Also
    --------
    mean_covariance
    """
    X = _deprecate_covmats(covmats, X)
    return np.average(X, axis=0, weights=sample_weight)


def mean_harmonic(X=None, sample_weight=None, covmats=None):
    r"""Harmonic mean of invertible matrices.

    .. math::
        \mathbf{M} = \left( \sum_i w_i \ {\mathbf{X}_i}^{-1} \right)^{-1}

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of invertible matrices.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Harmonic mean.

    See Also
    --------
    mean_covariance
    """
    X = _deprecate_covmats(covmats, X)
    T = mean_euclid(np.linalg.inv(X), sample_weight=sample_weight)
    M = np.linalg.inv(T)
    return M


def mean_identity(X=None, sample_weight=None, covmats=None):
    r"""Identity matrix corresponding to the matrices dimension.

    .. math::
        \mathbf{M} = \mathbf{I}_n

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of square matrices.
    sample_weight : None
        Not used, here for compatibility with other means.

    Returns
    -------
    M : ndarray, shape (n, n)
        Identity matrix.

    See Also
    --------
    mean_covariance
    """
    X = _deprecate_covmats(covmats, X)
    M = np.eye(X.shape[-1])
    return M


def mean_kullback_sym(X=None, sample_weight=None, covmats=None):
    """Mean of SPD/HPD matrices according to Kullback-Leibler divergence.

    Symmetrized Kullback-Leibler mean is the geometric mean between the
    Euclidean and the harmonic means [1]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Symmetrized Kullback-Leibler mean.

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Symmetric positive-definite matrices: From geometry to applications
        and visualization
        <https://link.springer.com/chapter/10.1007/3-540-31272-2_17>`_
        M. Moakher and P. Batchelor. Visualization and Processing of Tensor
        Fields, pp. 285-298, 2006
    """
    X = _deprecate_covmats(covmats, X)
    M_euclid = mean_euclid(X, sample_weight=sample_weight)
    M_harmonic = mean_harmonic(X, sample_weight=sample_weight)
    M = geodesic_riemann(M_euclid, M_harmonic, 0.5)
    return M


def mean_logdet(X=None, tol=10e-5, maxiter=50, init=None, sample_weight=None,
                covmats=None):
    r"""Mean of SPD/HPD matrices according to the log-det metric.

    Log-det mean is obtained by an iterative procedure where the update is:

    .. math::
        \mathbf{M} = \left( \sum_i w_i \ \left( 0.5 \mathbf{M}
                     + 0.5 \mathbf{X}_i \right)^{-1} \right)^{-1}

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    tol : float, default=10e-5
        The tolerance to stop the gradient descent.
    maxiter : int, default=50
        The maximum number of iterations.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None, the weighted Euclidean mean is used.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Log-det mean.

    See Also
    --------
    mean_covariance
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, _, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)
    if init is None:
        M = mean_euclid(X, sample_weight=sample_weight)
    else:
        M = init

    crit = np.finfo(np.float64).max
    for _ in range(maxiter):
        invX = np.linalg.inv(0.5 * X + 0.5 * M)
        J = np.einsum('a,abc->bc', sample_weight, invX)
        Mnew = np.linalg.inv(J)

        crit = np.linalg.norm(Mnew - M, ord='fro')
        M = Mnew
        if crit <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    return M


def mean_logeuclid(X=None, sample_weight=None, covmats=None):
    r"""Mean of SPD/HPD matrices according to the log-Euclidean metric.

    Log-Euclidean mean is [1]_:

    .. math::
        \mathbf{M} = \exp{ \left( \sum_i w_i \ \log{\mathbf{X}_i} \right) }

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Log-Euclidean mean.

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Geometric means in a novel vector space structure on symmetric
        positive-definite matrices
        <https://epubs.siam.org/doi/abs/10.1137/050637996?journalCode=sjmael>`_
        V. Arsigny, P. Fillard, X. Pennec, and N. Ayache. SIAM Journal on
        Matrix Analysis and Applications. Volume 29, Issue 1 (2007).
    """
    X = _deprecate_covmats(covmats, X)
    M = expm(mean_euclid(logm(X), sample_weight=sample_weight))
    return M


def mean_power(X=None, p=None, *, sample_weight=None, zeta=10e-10, maxiter=100,
               covmats=None):
    r"""Power mean of SPD/HPD matrices.

    Power mean of order p is the solution of [1]_ [2]_:

    .. math::
        \mathbf{M} = \sum_i w_i \ \mathbf{M} \sharp_p \mathbf{X}_i

    where :math:`\mathbf{A} \sharp_p \mathbf{B}` is the geodesic between
    matrices :math:`\mathbf{A}` and :math:`\mathbf{B}`.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    p : float
        Exponent, in [-1,+1]. For p=0, it returns
        :func:`pyriemann.utils.mean.mean_riemann`.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.
    zeta : float, default=10e-10
        Stopping criterion.
    maxiter : int, default=100
        The maximum number of iterations.

    Returns
    -------
    M : ndarray, shape (n, n)
        Power mean.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Matrix Power means and the Karcher mean
        <https://www.sciencedirect.com/science/article/pii/S0022123611004101>`_
        Y. Lim and M. Palfia. Journal of Functional Analysis, Volume 262,
        Issue 4, 15 February 2012, Pages 1498-1514.
    .. [2] `Fixed Point Algorithms for Estimating Power Means of Positive
        Definite Matrices
        <https://hal.archives-ouvertes.fr/hal-01500514>`_
        M. Congedo, A. Barachant, and R. Bhatia. IEEE Transactions on Signal
        Processing, Volume 65, Issue 9, pp.2211-2220, May 2017
    """
    X = _deprecate_covmats(covmats, X)
    if p is None:
        raise ValueError("Input p can not be None")
    if not isinstance(p, (int, float)):
        raise ValueError("Power mean only defined for a scalar exponent")
    if p < -1 or 1 < p:
        raise ValueError("Exponent must be in [-1,+1]")

    if p == 1:
        return mean_euclid(X, sample_weight=sample_weight)
    elif p == 0:
        return mean_riemann(X, sample_weight=sample_weight)
    elif p == -1:
        return mean_harmonic(X, sample_weight=sample_weight)

    n_matrices, n, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)
    phi = 0.375 / np.abs(p)

    G = np.einsum('a,abc->bc', sample_weight, powm(X, p))
    if p > 0:
        K = invsqrtm(G)
    else:
        K = sqrtm(G)

    eye_n, sqrt_n = np.eye(n), np.sqrt(n)
    crit = 10 * zeta
    for _ in range(maxiter):
        H = np.einsum(
            'a,abc->bc',
            sample_weight,
            powm(K @ powm(X, np.sign(p)) @ K.conj().T, np.abs(p))
        )
        K = powm(H, -phi) @ K

        crit = np.linalg.norm(H - eye_n) / sqrt_n
        if crit <= zeta:
            break
    else:
        warnings.warn("Convergence not reached")

    M = K.conj().T @ K
    if p > 0:
        M = np.linalg.inv(M)

    return M


def mean_riemann(X=None, tol=10e-9, maxiter=50, init=None, sample_weight=None,
                 covmats=None):
    r"""Mean of SPD/HPD matrices according to the Riemannian metric.

    The affine-invariant Riemannian mean minimizes the sum of squared
    affine-invariant Riemannian distances :math:`d_R` to all SPD/HPD matrices
    [1]_ [2]_:

    .. math::
         \arg \min_{\mathbf{M}} \sum_i w_i \ d_R (\mathbf{M}, \mathbf{X}_i)^2

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    tol : float, default=10e-9
        The tolerance to stop the gradient descent.
    maxiter : int, default=50
        The maximum number of iterations.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None, the weighted Euclidean mean is used.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Affine-invariant Riemannian mean.

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Principal geodesic analysis for the study of nonlinear statistics
        of shape
        <https://ieeexplore.ieee.org/document/1318725>`_
        P.T. Fletcher, C. Lu, S. M. Pizer, S. Joshi.
        IEEE Trans Med Imaging, 2004, 23(8), pp. 995-1005
    .. [2] `A differential geometric approach to the geometric mean of
        symmetric positive-definite matrices
        <https://epubs.siam.org/doi/10.1137/S0895479803436937>`_
        M. Moakher. SIAM J Matrix Anal Appl, 2005, 26 (3), pp. 735-747
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, _, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)
    if init is None:
        M = mean_euclid(X, sample_weight=sample_weight)
    else:
        M = init

    nu = 1.0
    tau = np.finfo(np.float64).max
    crit = np.finfo(np.float64).max
    for _ in range(maxiter):
        M12, Mm12 = sqrtm(M), invsqrtm(M)
        J = np.einsum('a,abc->bc', sample_weight, logm(Mm12 @ X @ Mm12))
        M = M12 @ expm(nu * J) @ M12

        crit = np.linalg.norm(J, ord='fro')
        h = nu * crit
        if h < tau:
            nu = 0.95 * nu
            tau = h
        else:
            nu = 0.5 * nu
        if crit <= tol or nu <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    return M


def mean_wasserstein(X=None, tol=10e-4, maxiter=50, init=None,
                     sample_weight=None, covmats=None):
    r"""Mean of SPD/HPD matrices according to the Wasserstein metric.

    Wasserstein mean is obtained by an iterative procedure where the update is
    [1]_:

    .. math::
        \mathbf{K} = \left( \sum_i w_i \ \left( \mathbf{K} \mathbf{X}_i
                     \mathbf{K} \right)^{1/2} \right)^{1/2}

    with :math:`\mathbf{K} = \mathbf{M}^{1/2}`.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    tol : float, default=10e-4
        The tolerance to stop the gradient descent.
    maxiter : int, default=50
        The maximum number of iterations.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None the Euclidean mean is used.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Wasserstein mean.

    See Also
    --------
    mean_covariance

    References
    ----------
    .. [1] `Geometric Radar Processing based on Frechet distance: Information
        geometry versus Optimal Transport Theory
        <https://ieeexplore.ieee.org/document/6042179>`_
        F. Barbaresco. 12th International Radar Symposium (IRS), October 2011
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, _, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)
    if init is None:
        M = mean_euclid(X, sample_weight=sample_weight)
    else:
        M = init
    K = sqrtm(M)

    crit = np.finfo(np.float64).max
    for _ in range(maxiter):
        J = np.einsum('a,abc->bc', sample_weight, sqrtm(K @ X @ K))
        Knew = sqrtm(J)

        crit = np.linalg.norm(Knew - K, ord='fro')
        K = Knew
        if crit <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    M = K @ K
    return M


###############################################################################


mean_functions = {
    "ale": mean_ale,
    "alm": mean_alm,
    "euclid": mean_euclid,
    "harmonic": mean_harmonic,
    "identity": mean_identity,
    "kullback_sym": mean_kullback_sym,
    "logdet": mean_logdet,
    "logeuclid": mean_logeuclid,
    "riemann": mean_riemann,
    "wasserstein": mean_wasserstein,
}


def mean_covariance(X=None, metric="riemann", sample_weight=None, covmats=None,
                    **kwargs):
    """Mean of matrices according to a metric.

    Compute the mean of a set of matrices according to a metric [1]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of matrices.
    metric : string | callable, default="riemann"
        Metric for mean estimation, can be: "ale", "alm", "euclid", "harmonic",
        "identity", "kullback_sym", "logdet", "logeuclid", "riemann",
        "wasserstein", or a callable function.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.
    **kwargs : dict
        The keyword arguments passed to the sub function.

    Returns
    -------
    M : ndarray, shape (n, n)
        Mean of matrices.

    References
    ----------
    .. [1] `Review of Riemannian distances and divergences, applied to
        SSVEP-based BCI
        <https://hal.archives-ouvertes.fr/LISV/hal-03015762v1>`_
        S. Chevallier, E. K. Kalunga, Q. Barthélemy, E. Monacelli.
        Neuroinformatics, Springer, 2021, 19 (1), pp.93-106
    """
    X = _deprecate_covmats(covmats, X)
    mean_function = check_function(metric, mean_functions)
    M = mean_function(
        X,
        sample_weight=sample_weight,
        **kwargs,
    )
    return M


###############################################################################


def _get_mask_from_nan(X):
    nan_col = np.all(np.isnan(X), axis=0)
    nan_row = np.all(np.isnan(X), axis=1)
    if not np.array_equal(nan_col, nan_row):
        raise ValueError("NaN values are not symmetric.")
    nan_inds = np.where(nan_col)
    subX_ = np.delete(X, nan_inds, axis=0)
    subX = np.delete(subX_, nan_inds, axis=1)
    if np.any(np.isnan(subX)):
        raise ValueError("NaN values must fill rows and columns.")
    mask = np.delete(np.eye(X.shape[0]), nan_inds, axis=1)
    return mask


def _get_masks_from_nan(X):
    masks = []
    for i in range(len(X)):
        masks.append(_get_mask_from_nan(X[i]))
    return masks


def _apply_masks(X, masks):
    maskedX = []
    for i in range(len(X)):
        maskedX.append(masks[i].T @ X[i] @ masks[i])
    return maskedX


def maskedmean_riemann(X=None, masks=None, tol=10e-9, maxiter=100, init=None,
                       sample_weight=None, covmats=None):
    """Masked Riemannian mean of SPD/HPD matrices.

    Given masks defined as semi-orthogonal matrices, the masked Riemannian mean
    of SPD/HPD matrices is obtained with a gradient descent minimizing the sum
    of affine-invariant Riemannian distances between masked SPD/HPD matrices
    and the masked mean [1]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    masks : list of n_matrices ndarray of shape (n, n_i), \
            with different n_i, such that n_i <= n
        Masks, defined as semi-orthogonal matrices. See [1]_.
    tol : float, default=10e-9
        The tolerance to stop the gradient descent.
    maxiter : int, default=100
        The maximum number of iteration.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None, the Identity is used.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Masked Riemannian mean.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    mean_riemann
    mean_covariance

    References
    ----------
    .. [1] `Geodesically-convex optimization for averaging partially observed
        covariance matrices
        <https://hal.archives-ouvertes.fr/hal-02984423>`_
        F. Yger, S. Chevallier, Q. Barthélemy, and S. Sra. Asian Conference on
        Machine Learning (ACML), Nov 2020, Bangkok, Thailand. pp.417 - 432.
    """
    if masks is None:
        raise ValueError("Input masks can not be None")
    X = _deprecate_covmats(covmats, X)
    n_matrices, n, _ = X.shape
    sample_weight = check_weights(sample_weight, n_matrices)
    maskedX = _apply_masks(X, masks)
    if init is None:
        M = np.eye(n)
    else:
        M = init

    nu = 1.0
    tau = np.finfo(np.float64).max
    crit = np.finfo(np.float64).max
    for _ in range(maxiter):
        maskedM = _apply_masks(np.tile(M, (n_matrices, 1, 1)), masks)
        J = np.zeros((n, n), dtype=X.dtype)
        for i in range(n_matrices):
            M12, Mm12 = sqrtm(maskedM[i]), invsqrtm(maskedM[i])
            tmp = M12 @ logm(Mm12 @ maskedX[i] @ Mm12) @ M12
            J += sample_weight[i] * masks[i] @ tmp @ masks[i].T
        M12, Mm12 = sqrtm(M), invsqrtm(M)
        M = M12 @ expm(Mm12 @ (nu * J) @ Mm12) @ M12

        crit = np.linalg.norm(J, ord='fro')
        h = nu * crit
        if h < tau:
            nu = 0.95 * nu
            tau = h
        else:
            nu = 0.5 * nu
        if crit <= tol or nu <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    return M


def nanmean_riemann(X=None, tol=10e-9, maxiter=100, init=None,
                    sample_weight=None, covmats=None):
    """Riemannian NaN-mean of SPD/HPD matrices.

    The Riemannian NaN-mean is the masked Riemannian mean applied to SPD/HPD
    matrices potentially corrupted by symmetric NaN values [1]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices, corrupted by symmetric NaN values [1]_.
    tol : float, default=10e-9
        The tolerance to stop the gradient descent.
    maxiter : int, default=100
        The maximum number of iteration.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None, a regularized Euclidean NaN-mean is used.
    sample_weight : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, n)
        Riemannian NaN-mean.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    maskedmean_riemann
    mean_covariance

    References
    ----------
    .. [1] `Geodesically-convex optimization for averaging partially observed
        covariance matrices
        <https://hal.archives-ouvertes.fr/hal-02984423>`_
        F. Yger, S. Chevallier, Q. Barthélemy, and S. Sra. Asian Conference on
        Machine Learning (ACML), Nov 2020, Bangkok, Thailand. pp.417 - 432.
    """
    X = _deprecate_covmats(covmats, X)
    n_matrices, n, _ = X.shape
    if init is None:
        Minit = np.nanmean(X, axis=0) + 1e-6 * np.eye(n)
    else:
        Minit = init

    M = maskedmean_riemann(
        np.nan_to_num(X),  # avoid nan contamination in matmul
        _get_masks_from_nan(X),
        tol=tol,
        maxiter=maxiter,
        init=Minit,
        sample_weight=sample_weight
    )
    return M



import numpy as np


def encode_domains(X, y, domain):
    r"""Encode the domains of the matrices in the labels.

    We handle the possibility of having different domains for the datasets by
    extending the labels of the matrices and including this information to
    them. For instance, if we have a matrix X with class `left_hand` on the
    `domain_01` then its extended label will be `domain_01/left_hand`. Note
    that if the classes were integers at first, they will be converted to
    strings.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n_channels, n_channels)
        Set of SPD matrices.
    y : ndarray, shape (n_matrices,)
        Labels for each matrix.
    domain : ndarray, shape (n_matrices,)
        Domains for each matrix.

    Returns
    -------
    X_enc : ndarray, shape (n_matrices, n_channels, n_channels)
        The same set of SPD matrices given as input.
    y_enc : ndarray, shape (n_matrices,)
        Extended labels for each matrix.

    See Also
    --------
    decode_domains

    Notes
    -----
    .. versionadded:: 0.4
    """
    if len(y) != len(domain):
        raise ValueError("Input lengths don't match")

    y_enc = [str(d_) + '/' + str(y_) for (d_, y_) in zip(domain, y)]
    return X, np.array(y_enc)


def decode_domains(X_enc, y_enc):
    """Decode the domains of the matrices in the labels.

    We handle the possibility of having different domains for the datasets by
    encoding the domain information into the labels of the matrices. This
    method converts the data into its original form, with a separate data
    structure for labels and for domains.

    Parameters
    ----------
    X_enc : ndarray, shape (n_matrices, n_channels, n_channels)
        Set of SPD matrices.
    y_enc : ndarray, shape (n_matrices,)
        Extended labels for each matrix.

    Returns
    -------
    X : ndarray, shape (n_matrices, n_channels, n_channels)
        Set of SPD matrices.
    y : ndarray, shape (n_matrices,)
        Labels for each matrix.
    domain : ndarray, shape (n_matrices,)
        Domains for each matrix.

    See Also
    --------
    encode_domains

    Notes
    -----
    .. versionadded:: 0.4
    """
    y, domain = [], []
    for y_enc_ in y_enc:
        y_dec_ = y_enc_.split('/')
        domain.append(y_dec_[-2])
        y.append(y_dec_[-1])
    return X_enc, np.array(y), np.array(domain)


class TLSplitter():
    """Class for handling the cross-validation splits of multi-domain data.

    This is a wrapper to sklearn's cross-validation iterators [1]_ which
    ensures the handling of domain information with the data points. In fact,
    the data from source domain is always fully available in the training
    partition whereas the random splits are done on the data points from the
    target domain.

    Parameters
    ----------
    target_domain : str
        Domain considered as target.
    cv : None | BaseCrossValidator | BaseShuffleSplit, default=None
        An instance of a cross validation iterator from sklearn.

    References
    ----------
    .. [1] https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators

    Notes
    -----
    .. versionadded:: 0.4
    """  # noqa
    def __init__(self, target_domain, cv):

        self.target_domain = target_domain
        self.cv = cv

    def split(self, X, y):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_channels)
            Set of SPD matrices.
        y : ndarray, shape (n_matrices,)
            Extended labels for each matrix.

        Yields
        ------
        train : ndarray
            The training set indices for that split.
        test : ndarray
            The testing set indices for that split.
        """

        # decode the domains of the data points
        X, y, domain = decode_domains(X, y)

        # indentify the indices of the target dataset
        idx_source = np.where(domain != self.target_domain)[0]
        idx_target = np.where(domain == self.target_domain)[0]
        y_target = y[idx_target]

        # index of training-split for the target data points
        ss_target = self.cv.split(idx_target, y_target)
        for train_sub_idx_target, test_sub_idx_target in ss_target:
            train_idx = np.concatenate(
                [idx_source, idx_target[train_sub_idx_target]])
            test_idx = idx_target[test_sub_idx_target]
            yield train_idx, test_idx

    def get_n_splits(self, X=None, y=None):
        """Returns the number of splitting iterations in the cross-validator.

        Parameters
        ----------
        X : object
            Ignored, exists for compatibility.
        y : object
            Ignored, exists for compatibility.

        Returns
        -------
        n_splits : int
            Returns the number of splitting iterations in the cross-validator.
        """
        return self.cv.n_splits



import numpy as np
import pytest
from pytest import approx

from pyriemann.utils.median import (
    median_euclid,
    median_riemann,
)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("median", [median_euclid, median_riemann])
def test_median_shape(kind, median, get_mats):
    """Test the shape of median"""
    n_matrices, n_channels = 3, 4
    mats = get_mats(n_matrices, n_channels, kind)
    C = median(mats)
    assert C.shape == (n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("median", [median_euclid, median_riemann])
def test_median_shape_with_init(kind, median, get_mats):
    """Test the shape of median with init"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, kind)
    C = median(mats, init=mats[0])
    assert C.shape == (n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("median", [median_euclid, median_riemann])
def test_median_weight_zero(kind, median, get_mats):
    """Setting one weight to almost 0 it's almost like not passing the mat"""
    n_matrices, n_channels, w_val = 5, 3, 2
    mats = get_mats(n_matrices, n_channels, kind)
    w = w_val * np.ones(n_matrices)
    C = median(mats[1:], weights=w[1:])
    w[0] = 1e-12
    Cw = median(mats, weights=w)
    assert C == approx(Cw, rel=1e-6, abs=1e-8)


@pytest.mark.parametrize("median", [median_euclid, median_riemann])
def test_median_warning_convergence(median, get_mats):
    """Test warning for convergence not reached"""
    n_matrices, n_channels = 3, 2
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.warns(UserWarning):
        median(mats, maxiter=0)


@pytest.mark.parametrize("n_values", [3, 5, 7])
def test_median_euclid_1d(n_values, rndstate):
    """Compare geometric Euclidean median to marginal median in 1D"""
    values = 100 * rndstate.randn(n_values)
    np_med = np.median(values)
    py_med = median_euclid(values[..., np.newaxis, np.newaxis])[0, 0]
    assert np_med == approx(py_med)


@pytest.mark.parametrize("complex_valued", [True, False])
def test_median_euclid(rndstate, complex_valued):
    """Test the Euclidean median for generic matrices"""
    n_matrices, n_dim0, n_dim1 = 10, 3, 4
    mats = rndstate.randn(n_matrices, n_dim0, n_dim1)
    if complex_valued:
        mats = mats + 1j * rndstate.randn(n_matrices, n_dim0, n_dim1)
    assert median_euclid(mats).shape == (n_dim0, n_dim1)


@pytest.mark.parametrize("step_size", [0, 2.5])
def test_median_riemann_stepsize_error(step_size, get_mats):
    n_matrices, n_channels = 1, 2
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        median_riemann(mats, step_size=step_size)



import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from pyriemann.utils.base import (
    expm,
    invsqrtm,
    logm,
    powm,
    sqrtm,
    nearest_sym_pos_def,
)
from pyriemann.utils.mean import mean_riemann
from pyriemann.utils.test import is_pos_def, is_sym_pos_def


n_channels = 3


def test_expm():
    """Test matrix exponential"""
    C = 2 * np.eye(n_channels)
    Ctrue = np.exp(2) * np.eye(n_channels)
    assert_array_almost_equal(expm(C), Ctrue, decimal=10)


def test_invsqrtm():
    """Test matrix inverse square root"""
    C = 2 * np.eye(n_channels)
    Ctrue = (1.0 / np.sqrt(2)) * np.eye(n_channels)
    assert_array_almost_equal(invsqrtm(C), Ctrue, decimal=10)


def test_logm():
    """Test matrix logarithm"""
    C = 2 * np.eye(n_channels)
    Ctrue = np.log(2) * np.eye(n_channels)
    assert_array_almost_equal(logm(C), Ctrue, decimal=10)


def test_powm():
    """Test matrix power"""
    C = 2 * np.eye(n_channels)
    Ctrue = (2 ** 0.3) * np.eye(n_channels)
    assert_array_almost_equal(powm(C, 0.3), Ctrue, decimal=10)


def test_sqrtm():
    """Test matrix square root"""
    C = 2 * np.eye(n_channels)
    Ctrue = np.sqrt(2) * np.eye(n_channels)
    assert_array_almost_equal(sqrtm(C), Ctrue, decimal=10)

    C = np.array([[1, -1j], [1j, 1]])
    Ctrue = np.sqrt(2) / 2 * C
    assert_array_almost_equal(sqrtm(C), Ctrue, decimal=10)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("funm", [expm, invsqrtm, logm, powm, sqrtm])
def test_funm_all(kind, funm, get_mats):
    n_matrices, n_dim = 10, 3
    C = get_mats(n_matrices, n_dim, kind)
    if funm is powm:
        D = funm(C, 0.42)
    else:
        D = funm(C)
    assert D.shape == (n_matrices, n_dim, n_dim)


def test_funm_error():
    with pytest.raises(ValueError):
        sqrtm(np.ones(5))
    with pytest.raises(ValueError):
        invsqrtm(5.1)
    with pytest.raises(ValueError):
        logm([5.2])


@pytest.mark.parametrize("funm", [expm, invsqrtm, logm, powm, sqrtm])
def test_funm_ndarray(funm):
    def test(funm, C):
        if funm == powm:
            D = funm(C, 0.2)
        else:
            D = funm(C)
        assert C.shape == D.shape

    n_matrices = 6
    C_3d = np.asarray([np.eye(n_channels) for _ in range(n_matrices)])
    test(funm, C_3d)

    n_sets = 5
    C_4d = np.asarray([C_3d for _ in range(n_sets)])
    test(funm, C_4d)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_funm_properties(get_mats, kind):
    n_matrices, n_dim = 10, 3
    C = get_mats(n_matrices, n_dim, kind)
    invC = np.linalg.inv(C)

    # expm and logm
    eC, lC = expm(C), logm(C)
    assert_array_almost_equal(eC.conj(), expm(C.conj()), decimal=10)
    assert_array_almost_equal(
        np.linalg.det(eC),
        np.exp(np.trace(C, axis1=-2, axis2=-1)),
        decimal=10,
    )
    assert_array_almost_equal(logm(eC), C, decimal=10)
    assert_array_almost_equal(expm(lC), C, decimal=10)
    assert_array_almost_equal(expm(-lC), invC, decimal=10)

    # invsqrtm
    isC = invsqrtm(C)
    Eye = np.repeat(np.eye(n_dim)[np.newaxis, :, :], n_matrices, axis=0)
    assert_array_almost_equal(isC @ C @ isC, Eye, decimal=10)
    assert_array_almost_equal(isC @ isC, invC, decimal=10)

    # sqrtm
    sC = sqrtm(C)
    assert_array_almost_equal(
        np.swapaxes(sC.conj(), -2, -1) @ sC,
        C,
        decimal=10,
    )
    assert_array_almost_equal(isC @ C @ isC, Eye, decimal=10)

    # powm
    assert_array_almost_equal(powm(C, 0.5), sC, decimal=10)
    assert_array_almost_equal(powm(C, -0.5), isC, decimal=10)
    alpha = 0.3
    assert_array_almost_equal(
        powm(C, alpha=alpha),
        expm(alpha * logm(C)),
        decimal=10,
    )


def test_check_raise():
    """Test check SPD matrices"""
    C = 2 * np.ones((10, n_channels, n_channels))
    # This is an indirect check, the riemannian mean must crash when the
    # matrices are not SPD.
    with pytest.warns(RuntimeWarning):
        with pytest.raises(ValueError):
            mean_riemann(C)


def test_nearest_sym_pos_def(get_mats):
    n_matrices = 3
    mats = get_mats(n_matrices, n_channels, "spd")
    D = mats.diagonal(axis1=1, axis2=2)
    psd = np.array([mat - np.diag(d) for mat, d in zip(mats, D)])

    assert not is_pos_def(psd)
    assert is_sym_pos_def(nearest_sym_pos_def(mats))
    assert is_sym_pos_def(nearest_sym_pos_def(psd))



import numpy as np
import pytest

from pyriemann.utils.test import (
    is_square, is_sym, is_skew_sym, is_hankel,
    is_real, is_real_type, is_hermitian,
    is_pos_def, is_pos_semi_def,
    is_sym_pos_def, is_sym_pos_semi_def,
    is_herm_pos_def, is_herm_pos_semi_def,
)


n = 3


def test_is_square():
    assert is_square(np.eye(n))
    assert not is_square(np.ones((n, n + 1)))


def test_is_sym(rndstate):
    assert is_sym(np.eye(n))

    A = rndstate.randn(n, n)
    assert is_sym(A + A.T)
    assert not is_sym(-A + A.T)
    assert not is_sym(np.ones((n, n + 1)))


def test_is_skew_sym(rndstate):
    A = rndstate.randn(n, n)
    assert is_skew_sym(A - A.T)
    assert not is_skew_sym(np.eye(n))
    assert not is_skew_sym(A + A.T)
    assert not is_skew_sym(np.ones((n, n + 1)))


def test_is_hankel():
    assert is_hankel(np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5]]))
    assert not is_hankel(np.array([[1, 2, 3], [1, 3, 4], [3, 4, 5]]))
    assert not is_hankel(np.array([[1, 2, 3], [2, 3, 3], [3, 4, 5]]))


def test_is_real(rndstate):
    A = rndstate.randn(n, n + 2)
    assert is_real(A)

    B = np.zeros((n, n + 2), dtype=complex)
    B.real = A
    assert is_real(B)
    B.imag = A / 1000.0
    assert not is_real(B)


def test_is_real_type(rndstate):
    A = np.zeros((n, n + 1))
    assert is_real_type(A)

    B = np.zeros((n, n + 2), dtype=complex)
    assert not is_real_type(B)


def test_is_hermitian(rndstate):
    A = rndstate.randn(n, n)
    B = np.zeros((n, n), dtype=complex)
    B.real, B.imag = A + A.T, A - A.T
    assert is_hermitian(B)
    assert not is_hermitian(np.ones((n, n)) + 1j * np.ones((n, n)))


@pytest.mark.parametrize("fast_mode", [True, False])
def test_is_pos_def(rndstate, fast_mode):
    assert is_pos_def(np.eye(n), fast_mode=fast_mode)

    A = rndstate.randn(n, 100)
    assert is_pos_def(A @ A.T + 1e-6 * np.eye(n), fast_mode=fast_mode)
    assert not is_pos_def(-A @ A.T, fast_mode=fast_mode)
    assert not is_pos_def(np.ones((n, n + 1)), fast_mode=fast_mode)


def test_is_pos_semi_def(rndstate):
    assert is_pos_semi_def(np.eye(n))

    A = rndstate.randn(n, 100)
    assert is_pos_semi_def(A @ A.T)
    assert not is_pos_semi_def(-A @ A.T)
    assert not is_pos_semi_def(np.ones((n, n + 1)))


def test_is_sym_pos_def(rndstate):
    assert is_sym_pos_def(np.eye(n))

    A = rndstate.randn(n, 100)
    assert is_sym_pos_def(A @ A.T + 1e-6 * np.eye(n))
    assert not is_sym_pos_def(-A @ A.T)
    assert not is_sym_pos_def(np.ones((n, n + 1)))

    B = A - np.mean(A, axis=0)
    assert not is_sym_pos_def(B @ B.T, tol=1e-9)


def test_is_sym_pos_semi_def(rndstate):
    assert is_sym_pos_semi_def(np.eye(n))

    A = rndstate.randn(n, 100)
    assert is_sym_pos_semi_def(A @ A.T)
    assert not is_sym_pos_semi_def(-A @ A.T)
    assert not is_sym_pos_semi_def(np.ones((n, n + 1)))

    B = A - np.mean(A, axis=0)
    assert is_sym_pos_semi_def(B @ B.T)


def test_is_herm_pos_def(rndstate):
    assert is_herm_pos_def(np.eye(n))

    A = rndstate.randn(n, 100) + 1j * rndstate.randn(n, 100)
    assert is_herm_pos_def(A @ A.conj().T + 1e-6 * np.eye(n))
    assert not is_herm_pos_def(A @ A.T)  # pseudo-covariance is not HPD
    assert not is_herm_pos_def(-A @ A.conj().T)
    assert not is_herm_pos_def(np.ones((n, n + 1)))

    B = A - np.mean(A, axis=0)
    assert not is_herm_pos_def(B @ B.conj().T, tol=1e-9)


def test_is_herm_pos_semi_def(rndstate):
    assert is_herm_pos_semi_def(np.eye(n))

    A = rndstate.randn(n, 100) + 1j * rndstate.randn(n, 100)
    assert is_herm_pos_semi_def(A @ A.conj().T)
    assert not is_herm_pos_semi_def(A @ A.T)  # pseudo-covariance is not HPSD
    assert not is_herm_pos_semi_def(-A @ A.conj().T)
    assert not is_herm_pos_semi_def(np.ones((n, n + 1)))

    B = A - np.mean(A, axis=0)
    assert is_herm_pos_semi_def(B @ B.conj().T + 1e-9 * np.eye(n))



import pickle

import numpy as np
from numpy.testing import assert_array_equal
import pytest
from pytest import approx
from scipy.stats import mode
from sklearn.dummy import DummyClassifier

from conftest import get_distances, get_means, get_metrics
from pyriemann.estimation import Covariances
from pyriemann.utils.kernel import kernel
from pyriemann.utils.mean import mean_covariance
from pyriemann.classification import (
    _mode_2d,
    MDM,
    FgMDM,
    KNearestNeighbor,
    TSclassifier,
    SVC,
    MeanField,
    class_distinctiveness,
)

rclf = [MDM, FgMDM, KNearestNeighbor, TSclassifier, SVC, MeanField]


@pytest.mark.parametrize(
    "X, axis, expected",
    [
        (
            np.array([[0, 5], [1, 7], [0, 6], [2, 7]]),
            0,
            np.array([0, 7]),
        ),
        (
            np.array([[0, 1, 2, 1, 1, 0], [7, 5, 7, 6, 7, 6]]),
            1,
            np.array([1, 7]),
        ),
        (
            np.array([['a', 'b', 'a', 'c', 'a'], ['d', 'e', 'f', 'e', 'e']]),
            1,
            np.array(['a', 'e']),
        ),
    ],
)
def test_mode(X, axis, expected):
    actual = _mode_2d(X, axis=axis)
    assert_array_equal(actual, expected)

    if np.issubdtype(X.dtype, np.number):
        sp, _ = mode(X, axis=axis)
        assert_array_equal(actual, sp.ravel())


@pytest.mark.parametrize("n_classes", [2, 3])
@pytest.mark.parametrize("classif", rclf)
def test_classifier(n_classes, classif, get_mats, get_labels):
    if n_classes == 2:
        n_matrices, n_channels = 6, 3
    else:
        assert n_classes == 3
        n_matrices, n_channels = 9, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    clf_predict(classif, covmats, labels)
    clf_fit_independence(classif, covmats, labels)
    clf_predict_proba(classif, covmats, labels)
    clf_score(classif, covmats, labels)
    clf_populate_classes(classif, covmats, labels)
    if classif in (MDM, KNearestNeighbor, MeanField):
        clf_fitpredict(classif, covmats, labels)
    if classif in (MDM, FgMDM):
        clf_transform(classif, covmats, labels)
    if hasattr(classif(), 'n_jobs'):
        clf_jobs(classif, covmats, labels)
    if hasattr(classif(), 'tsupdate'):
        clf_tsupdate(classif, covmats, labels)


def clf_predict(classif, covmats, labels):
    n_matrices = len(labels)
    clf = classif()
    clf.fit(covmats, labels)
    predicted = clf.predict(covmats)
    assert predicted.shape == (n_matrices,)


def clf_predict_proba(classif, covmats, labels):
    n_matrices = len(labels)
    n_classes = len(np.unique(labels))
    clf = classif()
    if hasattr(clf, 'probability'):
        clf.set_params(**{'probability': True})
    clf.fit(covmats, labels)
    probabilities = clf.predict_proba(covmats)
    assert probabilities.shape == (n_matrices, n_classes)
    assert probabilities.sum(axis=1) == approx(np.ones(n_matrices))


def clf_fitpredict(classif, covmats, labels):
    clf = classif()
    clf.fit_predict(covmats, labels)
    assert_array_equal(clf.classes_, np.unique(labels))


def clf_score(classif, covmats, labels):
    clf = classif()
    clf.fit(covmats, labels)
    clf.score(covmats, labels)


def clf_transform(classif, covmats, labels):
    n_matrices = len(labels)
    n_classes = len(np.unique(labels))
    clf = classif()
    clf.fit(covmats, labels)
    transformed = clf.transform(covmats)
    assert transformed.shape == (n_matrices, n_classes)


def clf_fit_independence(classif, covmats, labels):
    clf = classif()
    clf.fit(covmats, labels).predict(covmats)
    # retraining with different size should erase previous fit
    new_covmats = covmats[:, :-1, :-1]
    clf.fit(new_covmats, labels).predict(new_covmats)


def clf_jobs(classif, covmats, labels):
    clf = classif(n_jobs=2)
    clf.fit(covmats, labels)
    clf.predict(covmats)


def clf_populate_classes(classif, covmats, labels):
    clf = classif()
    clf.fit(covmats, labels)
    assert_array_equal(clf.classes_, np.unique(labels))


def clf_tsupdate(classif, covmats, labels):
    clf = classif(tsupdate=True)
    clf.fit(covmats, labels).predict(covmats)


@pytest.mark.parametrize("classif", rclf)
@pytest.mark.parametrize("mean", ["faulty", 42])
@pytest.mark.parametrize("dist", ["not_real", 27])
def test_metric_dict_error(classif, mean, dist, get_mats, get_labels):
    n_matrices, n_channels, n_classes = 6, 3, 2
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = classif(metric={"mean": mean, "distance": dist})
    with pytest.raises((TypeError, KeyError, ValueError)):
        clf.fit(covmats, labels).predict(covmats)


@pytest.mark.parametrize("classif", rclf)
@pytest.mark.parametrize("metric", [42, "faulty", {"foo": "bar"}])
def test_metric_errors(classif, metric, get_mats, get_labels):
    n_matrices, n_channels, n_classes = 6, 3, 2
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = classif(metric=metric)
    with pytest.raises((TypeError, KeyError, ValueError)):
        clf.fit(covmats, labels).predict(covmats)


@pytest.mark.parametrize("classif", rclf)
@pytest.mark.parametrize("metric", get_metrics())
def test_metric_str(classif, metric, get_mats, get_labels):
    n_matrices, n_channels, n_classes = 6, 3, 2
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = classif(metric=metric)

    if classif in [SVC, FgMDM, TSclassifier] \
            and metric not in ['riemann', 'euclid', 'logeuclid']:
        with pytest.raises((KeyError, ValueError)):
            clf.fit(covmats, labels).predict(covmats)
    else:
        clf.fit(covmats, labels).predict(covmats)


@pytest.mark.parametrize("metric_mean", get_means())
@pytest.mark.parametrize("metric_dist", get_distances())
@pytest.mark.parametrize("n_classes", [2, 3, 4])
def test_metric_mdm(metric_mean, metric_dist, n_classes, get_mats, get_labels):
    n_matrices, n_channels = 4 * n_classes, 3
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = MDM(metric={"mean": metric_mean, "distance": metric_dist})
    clf.fit(covmats, labels).predict(covmats)
    assert clf.classes_.shape == (n_classes,)
    assert clf.covmeans_.shape == (n_classes, n_channels, n_channels)


@pytest.mark.parametrize("metric_mean", get_means())
@pytest.mark.parametrize("metric_dist", get_distances())
@pytest.mark.parametrize("metric_map", ["euclid", "logeuclid", "riemann"])
def test_metric_fgmdm(metric_mean, metric_dist, metric_map, get_mats,
                      get_labels):
    n_matrices, n_channels, n_classes = 4, 3, 2
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = FgMDM(metric={
        "mean": metric_mean,
        "distance": metric_dist,
        "map": metric_map
    })
    clf.fit(covmats, labels).predict(covmats)


@pytest.mark.parametrize("metric_mean", get_means())
@pytest.mark.parametrize("metric_map", ["euclid", "logeuclid", "riemann"])
def test_metric_tsclassifier(metric_mean, metric_map, get_mats, get_labels):
    n_matrices, n_channels, n_classes = 4, 3, 2
    labels = get_labels(n_matrices, n_classes)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = TSclassifier(metric={"mean": metric_mean, "map": metric_map})
    clf.fit(covmats, labels).predict(covmats)


def test_1nn(get_mats, get_labels):
    """Test KNearestNeighbor with K=1"""
    n_matrices, n_channels, n_classes = 9, 3, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    knn = KNearestNeighbor(1, metric="riemann")
    knn.fit(covmats, labels)
    preds = knn.predict(covmats)
    assert_array_equal(labels, preds)


def test_tsclassifier_fit(get_mats, get_labels):
    """Test TS Classifier"""
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    clf = TSclassifier(clf=DummyClassifier())
    clf.fit(covmats, labels).predict(covmats)


def test_tsclassifier_clf_error(get_mats, get_labels):
    """Test TS if not Classifier"""
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    with pytest.raises(TypeError):
        TSclassifier(clf=Covariances()).fit(covmats, labels)


def test_svc_params():
    rsvc = SVC()
    assert rsvc.metric == 'riemann'

    rsvc.set_params(**{'metric': 'logeuclid'})
    assert rsvc.metric == 'logeuclid'

    rsvc.set_params(**{'max_iter': 501})
    assert rsvc.max_iter == 501


def test_svc_params_error(get_mats, get_labels):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    with pytest.raises(TypeError):
        SVC(C='hello').fit(covmats, labels)

    with pytest.raises(TypeError):
        SVC(foo=5)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_cref_metric(get_mats, get_labels, metric):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    Cref = mean_covariance(covmats, metric=metric)

    rsvc = SVC(Cref=Cref).fit(covmats, labels)
    rsvc_1 = SVC(Cref=None, metric=metric).fit(covmats, labels)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_cref_callable(get_mats, get_labels, metric):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    def Cref(X): return mean_covariance(X, metric=metric)

    rsvc = SVC(Cref=Cref).fit(covmats, labels)
    rsvc_1 = SVC(metric=metric).fit(covmats, labels)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)

    rsvc = SVC(Cref=Cref).fit(covmats, labels)
    rsvc.predict(covmats)
    rsvc_1 = SVC(metric=metric).fit(covmats, labels)
    rsvc_1.predict(covmats)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_cref_error(get_mats, get_labels, metric):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    def Cref(X, met): mean_covariance(X, metric=met)

    with pytest.raises(TypeError):
        SVC(Cref=Cref).fit(covmats, labels)

    Cref = metric

    with pytest.raises(TypeError):
        SVC(Cref=Cref).fit(covmats, labels)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_kernel_callable(get_mats, get_labels, metric):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    rsvc = SVC(kernel_fct=kernel, metric=metric).fit(covmats, labels)
    rsvc_1 = SVC(metric=metric).fit(covmats, labels)
    p1 = rsvc.predict(covmats[:-1])
    p2 = rsvc_1.predict(covmats[:-1])
    assert np.array_equal(p1, p2)

    def custom_kernel(X, Y, Cref, metric):
        return np.ones((len(X), len(Y)))
    SVC(kernel_fct=custom_kernel,
        metric=metric).fit(covmats, labels).predict(covmats[:-1])

    # check if pickleable
    pickle.dumps(rsvc)
    pickle.dumps(rsvc_1)


@pytest.mark.parametrize("kernel_fct", [None, "precomputed"])
@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_kernel_precomputed(get_mats, get_labels, kernel_fct, metric):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    SVC(kernel_fct=kernel_fct, metric=metric).fit(covmats, labels)


def test_svc_kernel_error(get_mats, get_labels):
    n_matrices, n_channels, n_classes = 4, 2, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)

    def custom_kernel(X, Y, Cref):
        return np.ones((len(X), len(Y)))
    with pytest.raises(TypeError):
        SVC(kernel_fct=custom_kernel, metric="euclid").fit(covmats, labels)

    custom_kernel = np.array([1, 2])
    with pytest.raises(TypeError):
        SVC(kernel_fct=custom_kernel, metric="riemann").fit(covmats, labels)


@pytest.mark.parametrize("method_label", ["sum_means", "inf_means"])
def test_meanfield(get_mats, get_labels, method_label):
    n_matrices, n_channels, n_classes = 6, 3, 2
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    mf = MeanField(method_label=method_label).fit(covmats, labels)
    pred = mf.predict(covmats)
    assert pred.shape == (n_matrices,)
    proba = mf.predict_proba(covmats)
    assert proba.shape == (n_matrices, n_classes)
    transf = mf.transform(covmats)
    assert transf.shape == (n_matrices, n_classes)


@pytest.mark.parametrize("n_classes", [1, 2, 3])
@pytest.mark.parametrize("metric_mean", get_means())
@pytest.mark.parametrize("metric_dist", get_distances())
@pytest.mark.parametrize("exponent", [1, 2])
def test_class_distinctiveness(get_mats, get_labels,
                               n_classes, metric_mean, metric_dist, exponent):
    """Test function for class distinctiveness measure for two class problem"""
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    if n_classes == 1:
        with pytest.raises(ValueError):
            class_distinctiveness(covmats, labels)
        return

    class_dis, num, denom = class_distinctiveness(
        covmats,
        labels,
        exponent,
        metric={"mean": metric_mean, "distance": metric_dist},
        return_num_denom=True
    )
    assert class_dis >= 0  # negative class_dis value
    assert num >= 0  # negative numerator value
    assert denom >= 0  # negative denominator value
    assert isinstance(class_dis, float), "Unexpected object of class_dis"
    assert isinstance(num, float), "Unexpected object of num"
    assert isinstance(denom, float), "Unexpected object of denum"



"""Geodesics for SPD/HPD matrices."""

from .base import sqrtm, invsqrtm, powm, logm, expm
from .utils import check_function


def geodesic_euclid(A, B, alpha=0.5):
    r"""Euclidean geodesic between matrices.

    The matrix at position :math:`\alpha` on the Euclidean geodesic
    between two matrices :math:`\mathbf{A}` and :math:`\mathbf{B}` is:

    .. math::
        \mathbf{C} = (1-\alpha) \mathbf{A} + \alpha \mathbf{B}

    :math:`\mathbf{C}` is equal to :math:`\mathbf{A}` if :math:`\alpha` = 0,
    and :math:`\mathbf{B}` if :math:`\alpha` = 1.

    Parameters
    ----------
    A : ndarray, shape (..., n, m)
        First matrices.
    B : ndarray, shape (..., n, m)
        Second matrices.
    alpha : float, default=0.5
        Position on the geodesic.

    Returns
    -------
    C : ndarray, shape (..., n, m)
        Matrices on the Euclidean geodesic.
    """
    return (1 - alpha) * A + alpha * B


def geodesic_logeuclid(A, B, alpha=0.5):
    r"""Log-Euclidean geodesic between SPD/HPD matrices.

    The matrix at position :math:`\alpha` on the Log-Euclidean geodesic
    between two SPD/HPD matrices :math:`\mathbf{A}` and :math:`\mathbf{B}` is:

    .. math::
        \mathbf{C} = \exp \left( (1-\alpha) \log(\mathbf{A})
                     + \alpha \log(\mathbf{B}) \right)

    :math:`\mathbf{C}` is equal to :math:`\mathbf{A}` if :math:`\alpha` = 0,
    and :math:`\mathbf{B}` if :math:`\alpha` = 1.

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices.
    alpha : float, default=0.5
        Position on the geodesic.

    Returns
    -------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices on the Log-Euclidean geodesic.
    """
    return expm((1 - alpha) * logm(A) + alpha * logm(B))


def geodesic_riemann(A, B, alpha=0.5):
    r"""Affine-invariant Riemannian geodesic between SPD/HPD matrices.

    The matrix at position :math:`\alpha` on the affine-invariant Riemannian
    geodesic between two SPD/HPD matrices :math:`\mathbf{A}` and
    :math:`\mathbf{B}` is:

    .. math::
        \mathbf{C} = \mathbf{A}^{1/2} \left( \mathbf{A}^{-1/2} \mathbf{B}
                     \mathbf{A}^{-1/2} \right)^\alpha \mathbf{A}^{1/2}

    :math:`\mathbf{C}` is equal to :math:`\mathbf{A}` if :math:`\alpha` = 0,
    and :math:`\mathbf{B}` if :math:`\alpha` = 1.

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices.
    alpha : float, default=0.5
        Position on the geodesic.

    Returns
    -------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices on the affine-invariant Riemannian geodesic.
    """
    sA, isA = sqrtm(A), invsqrtm(A)
    C = isA @ B @ isA
    D = powm(C, alpha)
    E = sA @ D @ sA
    return E


###############################################################################


geodesic_functions = {
    "euclid": geodesic_euclid,
    "logeuclid": geodesic_logeuclid,
    "riemann": geodesic_riemann,
}


def geodesic(A, B, alpha, metric="riemann"):
    """Geodesic between matrices according to a metric.

    Return the matrix at the position alpha on the geodesic between matrices
    A and B according to a metric.

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First matrices.
    B : ndarray, shape (..., n, n)
        Second matrices.
    alpha : float
        Position on the geodesic.
    metric : string | callable, default="riemann"
        Metric used for geodesic, can be: "euclid", "logeuclid", "riemann",
        or a callable function.

    Returns
    -------
    C : ndarray, shape (..., n, n)
        Matrices on the geodesic.
    """
    geodesic_function = check_function(metric, geodesic_functions)
    C = geodesic_function(A, B, alpha)
    return C



"""Tangent space for SPD/HPD matrices."""

import numpy as np

from .base import sqrtm, invsqrtm, logm, expm
from .mean import mean_covariance
from .utils import check_function


def _check_dimensions(X, Cref):
    n_1, n_2 = X.shape[-2:]
    n_3, n_4 = Cref.shape
    if not (n_1 == n_2 == n_3 == n_4):
        raise ValueError("Inputs have incompatible dimensions.")


def exp_map_euclid(X, Cref):
    r"""Project matrices back to manifold by Euclidean exponential map.

    The projection of a matrix :math:`\mathbf{X}` from tangent space
    to manifold with Euclidean exponential map
    according to a reference matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{original} = \mathbf{X} + \mathbf{C}_\text{ref}

    Parameters
    ----------
    X : ndarray, shape (..., n, m)
        Matrices in tangent space.
    Cref : ndarray, shape (n, m)
        Reference matrix.

    Returns
    -------
    X_original : ndarray, shape (..., n, m)
        Matrices in manifold.

    Notes
    -----
    .. versionadded:: 0.4
    """
    return X + Cref


def exp_map_logeuclid(X, Cref):
    r"""Project matrices back to manifold by Log-Euclidean exponential map.

    The projection of a matrix :math:`\mathbf{X}` from tangent space
    to SPD/HPD manifold with Log-Euclidean exponential map
    according to a reference SPD/HPD matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{original} =
        \exp(\mathbf{X} + \log(\mathbf{C}_\text{ref}))

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Matrices in tangent space.
    Cref : ndarray, shape (n, n)
        Reference SPD/HPD matrix.

    Returns
    -------
    X_original : ndarray, shape (..., n, n)
        Matrices in SPD/HPD manifold.

    Notes
    -----
    .. versionadded:: 0.4
    """
    return expm(X + logm(Cref))


def exp_map_riemann(X, Cref, Cm12=False):
    r"""Project matrices back to manifold by Riemannian exponential map.

    The projection of a matrix :math:`\mathbf{X}` from tangent space
    to SPD/HPD manifold with Riemannian exponential map
    according to a reference SPD/HPD matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{original} = \mathbf{C}_\text{ref}^{1/2}
        \exp(\mathbf{X}) \mathbf{C}_\text{ref}^{1/2}

    When Cm12=True, it returns the full Riemannian exponential map:

    .. math::
        \mathbf{X}_\text{original} = \mathbf{C}_\text{ref}^{1/2}
        \exp( \mathbf{C}_\text{ref}^{-1/2} \mathbf{X}
        \mathbf{C}_\text{ref}^{-1/2}) \mathbf{C}_\text{ref}^{1/2}

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Matrices in tangent space.
    Cref : ndarray, shape (n, n)
        Reference SPD/HPD matrix.
    Cm12 : bool, default=False
        If True, it returns the full Riemannian exponential map.

    Returns
    -------
    X_original : ndarray, shape (..., n, n)
        Matrices in SPD/HPD manifold.

    Notes
    -----
    .. versionadded:: 0.4
    """
    if Cm12:
        Cm12 = invsqrtm(Cref)
        X = Cm12 @ X @ Cm12
    C12 = sqrtm(Cref)
    return C12 @ expm(X) @ C12


def log_map_euclid(X, Cref):
    r"""Project matrices in tangent space by Euclidean logarithmic map.

    The projection of a matrix :math:`\mathbf{X}` from manifold
    to tangent space by Euclidean logarithmic map
    according to a reference matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{new} = \mathbf{X} - \mathbf{C}_\text{ref}

    Parameters
    ----------
    X : ndarray, shape (..., n, m)
        Matrices in manifold.
    Cref : ndarray, shape (n, m)
        Reference matrix.

    Returns
    -------
    X_new : ndarray, shape (..., n, m)
        Matrices projected in tangent space.

    Notes
    -----
    .. versionadded:: 0.4
    """
    return X - Cref


def log_map_logeuclid(X, Cref):
    r"""Project matrices in tangent space by Log-Euclidean logarithmic map.

    The projection of a matrix :math:`\mathbf{X}` from SPD/HPD manifold
    to tangent space by Log-Euclidean logarithmic map
    according to a SPD/HPD reference matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{new} = \log(\mathbf{X}) - \log(\mathbf{C}_\text{ref})

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Matrices in SPD/HPD manifold.
    Cref : ndarray, shape (n, n)
        Reference SPD matrix.

    Returns
    -------
    X_new : ndarray, shape (..., n, n)
        Matrices projected in tangent space.

    Notes
    -----
    .. versionadded:: 0.4
    """
    _check_dimensions(X, Cref)
    return logm(X) - logm(Cref)


def log_map_riemann(X, Cref, C12=False):
    r"""Project matrices in tangent space by Riemannian logarithmic map.

    The projection of a matrix :math:`\mathbf{X}` from SPD/HPD manifold
    to tangent space by Riemannian logarithmic map
    according to a SPD/HPD reference matrix :math:`\mathbf{C}_\text{ref}` is:

    .. math::
        \mathbf{X}_\text{new} = \log ( \mathbf{C}_\text{ref}^{-1/2}
        \mathbf{X} \mathbf{C}_\text{ref}^{-1/2})

    When C12=True, it returns the full Riemannian logarithmic map:

    .. math::
        \mathbf{X}_\text{new} = \mathbf{C}_\text{ref}^{1/2}
        \log( \mathbf{C}_\text{ref}^{-1/2} \mathbf{X}
        \mathbf{C}_\text{ref}^{-1/2}) \mathbf{C}_\text{ref}^{1/2}

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Matrices in SPD/HPD manifold.
    Cref : ndarray, shape (n, n)
        Reference SPD/HPD matrix.
    C12 : bool, default=False
        If True, it returns the full Riemannian logarithmic map.

    Returns
    -------
    X_new : ndarray, shape (..., n, n)
        Matrices projected in tangent space.

    Notes
    -----
    .. versionadded:: 0.4
    """
    _check_dimensions(X, Cref)
    Cm12 = invsqrtm(Cref)
    X_new = logm(Cm12 @ X @ Cm12)
    if C12:
        C12 = sqrtm(Cref)
        X_new = C12 @ X_new @ C12
    return X_new


def upper(X):
    r"""Return the weighted upper triangular part of matrices.

    This function computes the minimal representation of a matrix in tangent
    space [1]_: it keeps the upper triangular part of the symmetric/Hermitian
    matrix and vectorizes it by applying unity weight for diagonal elements and
    :math:`\sqrt{2}` weight for out-of-diagonal elements.

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Symmetric/Hermitian matrices.

    Returns
    -------
    T : ndarray, shape (..., n * (n + 1) / 2)
        Weighted upper triangular parts of symmetric/Hermitian matrices.

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] `Pedestrian detection via classification on Riemannian manifolds
        <https://ieeexplore.ieee.org/document/4479482>`_
        O. Tuzel, F. Porikli, and P. Meer. IEEE Transactions on Pattern
        Analysis and Machine Intelligence, Volume 30, Issue 10, October 2008.
    """
    n = X.shape[-1]
    if X.shape[-2] != n:
        raise ValueError("Matrices must be square")
    idx = np.triu_indices_from(np.empty((n, n)))
    coeffs = (np.sqrt(2) * np.triu(np.ones((n, n)), 1) + np.eye(n))[idx]
    T = coeffs * X[..., idx[0], idx[1]]
    return T


def unupper(T):
    """Inverse upper function.

    This function is the inverse of upper function: it reconstructs symmetric/
    Hermitian matrices from their weighted upper triangular parts.

    Parameters
    ----------
    T : ndarray, shape (..., n * (n + 1) / 2)
        Weighted upper triangular parts of symmetric/Hermitian matrices.

    Returns
    -------
    X : ndarray, shape (..., n, n)
        Symmetric/Hermitian matrices.

    See Also
    --------
    upper

    Notes
    -----
    .. versionadded:: 0.4
    """
    dims = T.shape
    n = int((np.sqrt(1 + 8 * dims[-1]) - 1) / 2)
    X = np.empty((*dims[:-1], n, n), dtype=T.dtype)
    idx = np.triu_indices_from(np.empty((n, n)))
    X[..., idx[0], idx[1]] = T
    idx = np.triu_indices_from(np.empty((n, n)), k=1)
    X[..., idx[0], idx[1]] /= np.sqrt(2)
    X[..., idx[1], idx[0]] = X[..., idx[0], idx[1]].conj()
    return X


log_map_functions = {
    "euclid": log_map_euclid,
    "logeuclid": log_map_logeuclid,
    "riemann": log_map_riemann,
}


def tangent_space(X, Cref, *, metric="riemann"):
    """Transform matrices into tangent vectors.

    Transform matrices into tangent vectors, according to a reference
    matrix Cref and to a specific logarithmic map.

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Matrices in manifold.
    Cref : ndarray, shape (n, n)
        Reference matrix.
    metric : string | callable, default="riemann"
        Metric used for logarithmic map, can be:
        "euclid", "logeuclid", "riemann", or a callable function.

    Returns
    -------
    T : ndarray, shape (..., n * (n + 1) / 2)
        Tangent vectors.

    See Also
    --------
    log_map_euclid
    log_map_logeuclid
    log_map_riemann
    upper
    """
    log_map_function = check_function(metric, log_map_functions)
    X_ = log_map_function(X, Cref)
    T = upper(X_)

    return T


exp_map_functions = {
    "euclid": exp_map_euclid,
    "logeuclid": exp_map_logeuclid,
    "riemann": exp_map_riemann,
}


def untangent_space(T, Cref, *, metric="riemann"):
    """Transform tangent vectors back to matrices.

    Transform tangent vectors back to matrices, according to a reference
    matrix Cref and to a specific exponential map.

    Parameters
    ----------
    T : ndarray, shape (..., n * (n + 1) / 2)
        Tangent vectors.
    Cref : ndarray, shape (n, n)
        Reference matrix.
    metric : string | callable, default="riemann"
        Metric used for exponential map, can be:
        "euclid", "logeuclid", "riemann", or a callable function.

    Returns
    -------
    X : ndarray, shape (..., n, n)
        Matrices in manifold.

    See Also
    --------
    unupper
    exp_map_euclid
    exp_map_logeuclid
    exp_map_riemann
    """
    X_ = unupper(T)
    exp_map_function = check_function(metric, exp_map_functions)
    X = exp_map_function(X_, Cref)

    return X


###############################################################################


# NOT IN API
def transport(Covs, Cref, metric="riemann"):
    """Parallel transport of a set of SPD matrices towards a reference matrix.

    Parameters
    ----------
    Covs : ndarray, shape (n_matrices, n, n)
        Set of SPD matrices.
    Cref : ndarray, shape (n, n)
        The reference SPD matrix.
    metric : string, default="riemann"
        The metric used for mean, can be: "euclid", "logeuclid", "riemann".

    Returns
    -------
    out : ndarray, shape (n_matrices, n, n)
        Set of transported SPD matrices.
    """
    C = mean_covariance(Covs, metric=metric)
    iC = invsqrtm(C)
    E = sqrtm(iC @ Cref @ iC)
    out = E @ Covs @ E.T
    return out



import os.path as op
import sys

from setuptools import find_packages, setup

version = None
with open(op.join("pyriemann", "_version.py"), "r") as fid:
    for line in (line.strip() for line in fid):
        if line.startswith("__version__"):
            version = line.split("=")[1].strip().strip("'")
            break
if version is None:
    raise RuntimeError("Could not determine version")

is_python3 = sys.version_info.major >= 3
kwargs_open = {"encoding": "utf8"} if is_python3 else {}
with open('README.md', 'r', **kwargs_open) as fid:
    long_description = fid.read()

setup(
    name="pyriemann",
    version=version,
    description="Biosignals classification with Riemannian geometry",
    url="https://pyriemann.readthedocs.io",
    author="Alexandre Barachant",
    author_email="alexandre.barachant@gmail.com",
    license="BSD (3-clause)",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": "https://pyriemann.readthedocs.io",
        "Source": "https://github.com/pyRiemann/pyRiemann",
        "Tracker": "https://github.com/pyRiemann/pyRiemann/issues/",
    },
    platforms="any",
    python_requires=">=3.8",
    install_requires=[
        "numpy!=1.24.0",
        "scipy",
        "scikit-learn>=0.24",
        "joblib",
        "matplotlib",
    ],
    extras_require={
        "docs": [
            "sphinx>=6.0.0,<=7.0.1",
            "sphinx-gallery",
            "sphinx-bootstrap_theme",
            "numpydoc",
            "mne",
            "seaborn",
            "pandas",
        ],
        "tests": [
            "pytest",
            "seaborn",
            "flake8"
        ],
    },
    zip_safe=False,
)



import warnings


class deprecated(object):
    """Mark a function or class as deprecated (decorator).

    Issue a warning when the function is called/the class is instantiated and
    adds a warning to the docstring.

    The optional extra argument will be appended to the deprecation message
    and the docstring. Note: to use this with the default value for extra, put
    in an empty of parentheses::

        >>> from pyriemann.utils import deprecated
        >>> deprecated()
        <pyriemann.utils.docs.deprecated object at ...>
        >>> @deprecated()
        ... def some_function(): pass


    Parameters
    ----------
    extra: string
        To be added to the deprecation messages.
    """

    # Borrowed from MNE:
    # https://mne.tools/stable/generated/mne.utils.deprecated.html

    def __init__(self, extra=""):
        self.extra = extra

    def __call__(self, obj):
        """Call.
        Parameters
        ----------
        obj : object
            Object to call.
        """
        if isinstance(obj, type):
            return self._decorate_class(obj)
        else:
            return self._decorate_fun(obj)

    def _decorate_class(self, cls):
        msg = "Class %s is deprecated" % cls.__name__
        if self.extra:
            msg += "; %s" % self.extra

        # FIXME: we should probably reset __new__ for full generality
        init = cls.__init__

        def deprecation_wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return init(*args, **kwargs)

        cls.__init__ = deprecation_wrapped

        deprecation_wrapped.__name__ = "__init__"
        deprecation_wrapped.__doc__ = self._update_doc(init.__doc__)
        deprecation_wrapped.deprecated_original = init

        return cls

    def _decorate_fun(self, fun):
        """Decorate function fun."""
        msg = "Function %s is deprecated" % fun.__name__
        if self.extra:
            msg += "; %s" % self.extra

        def deprecation_wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return fun(*args, **kwargs)

        deprecation_wrapped.__name__ = fun.__name__
        deprecation_wrapped.__dict__ = fun.__dict__
        deprecation_wrapped.__doc__ = self._update_doc(fun.__doc__)

        return deprecation_wrapped

    def _update_doc(self, olddoc):
        newdoc = ".. warning:: DEPRECATED"
        if self.extra:
            newdoc = "%s: %s" % (newdoc, self.extra)
        if olddoc:
            # Get the spacing right to avoid sphinx warnings
            n_space = 4
            for li, line in enumerate(olddoc.split("\n")):
                if li > 0 and len(line.strip()):
                    n_space = len(line) - len(line.lstrip())
                    break
            newdoc = "%s\n\n%s%s" % (newdoc, " " * n_space, olddoc)

        return newdoc



import numpy as np
import pytest

from conftest import requires_matplotlib
from pyriemann.utils.viz import (
    plot_embedding,
    plot_cospectra,
    plot_waveforms
)


@requires_matplotlib
def test_embedding(get_mats):
    """Test ."""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    plot_embedding(mats, y=None, metric="euclid")
    y = np.ones(mats.shape[0])
    plot_embedding(mats, y=y, metric="euclid")


@requires_matplotlib
def test_embedding_error_raise(get_mats):
    """Test ValueError for unknown embedding type."""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        plot_embedding(mats, y=None, metric="euclid", embd_type='foo')


@requires_matplotlib
def test_cospectra():
    """Test plot_cospectra"""
    n_freqs, n_channels = 16, 3
    cosp = np.random.randn(n_freqs, n_channels, n_channels)
    freqs = np.random.randn(n_freqs)
    plot_cospectra(cosp, freqs)


@requires_matplotlib
@pytest.mark.parametrize("display", ["all", "mean", "mean+/-std", "hist"])
def test_plot_waveforms(display):
    """Test plot_waveforms"""
    n_matrices, n_channels, n_times = 16, 3, 50
    X = np.random.randn(n_matrices, n_channels, n_times)
    plot_waveforms(X, display)
    plot_waveforms(X, display, times=np.arange(n_times))

    X = np.random.randn(n_matrices, 1, n_times)
    plot_waveforms(X, display)



"""Helpers for vizualization."""
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
import matplotlib.transforms as transforms
import numpy as np

from ..embedding import SpectralEmbedding, LocallyLinearEmbedding


def plot_embedding(X,
                   y=None,
                   *,
                   metric="riemann",
                   title="Embedding of covariances",
                   embd_type="Spectral",
                   normalize=True):
    """Plot 2D embedding of SPD matrices.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n_channels, n_channels)
        Set of SPD matrices.
    y : None | ndarray, shape (n_matrices,), default=None
        Labels for each matrix.
    metric : string, default="riemann"
        Metric used in the embedding. Can be {"riemann", "logeuclid",
        "euclid"} for Locally Linear Embedding, and {"riemann", "logeuclid",
        "euclid", "logdet", "kullback", "kullback_right", "kullback_sym"}
        for Spectral Embedding.
    title : str, default="Embedding of covariances"
        Title of figure.
    embd_type : {"Spectral", "LocallyLinear"}, default="Spectral"
        Embedding type.
    normalize : bool, default=True
        If True, the plot is normalized from -1 to +1.

    Returns
    -------
    fig : matplotlib figure
        Figure of embedding.

    Notes
    -----
    .. versionadded:: 0.2.6
    """
    if embd_type == "Spectral":
        lapl = SpectralEmbedding(n_components=2, metric=metric)
    elif embd_type == "LocallyLinear":
        lapl = LocallyLinearEmbedding(n_components=2,
                                      n_neighbors=X.shape[1],
                                      metric=metric)
    else:
        raise ValueError(f"Unknown embedding type {embd_type}. "
                         "Valid types are: 'Spectral', 'LocallyLinear'.")

    embd = lapl.fit_transform(X)

    if y is None:
        y = np.ones(embd.shape[0])

    fig, ax = plt.subplots(figsize=(7, 7), facecolor="white")
    for label in np.unique(y):
        idx = y == label
        ax.scatter(embd[idx, 0], embd[idx, 1], s=36)

    ax.set_xlabel(r"$\varphi_1$", fontsize=16)
    ax.set_ylabel(r"$\varphi_2$", fontsize=16)
    ax.set_title(f"{embd_type} {title}", fontsize=16)
    ax.grid(False)
    if normalize:
        ax.set_xticks([-1.0, -0.5, 0.0, +0.5, 1.0])
        ax.set_yticks([-1.0, -0.5, 0.0, +0.5, 1.0])
    ax.legend(list(np.unique(y)))

    return fig


def plot_cospectra(X, freqs, *, ylabels=None, title="Cospectra"):
    """Plot cospectral matrices.

    Parameters
    ----------
    X : ndarray, shape (n_freqs, n_channels, n_channels)
        Cospectral matrices.
    freqs : ndarray, shape (n_freqs,)
        The frequencies associated to cospectra.
    ylabels : list of str, default=None
        ylabels of figure.
    title : str, default="Cospectra"
        Title of figure.

    Returns
    -------
    fig : matplotlib figure
        Figure of cospectra.

    Notes
    -----
    .. versionadded:: 0.2.7
    """
    if X.ndim != 3:
        raise ValueError("Input X has not 3 dimensions")
    n_freqs, n_channels, _ = X.shape
    if freqs.shape != (n_freqs,):
        raise ValueError(
            "Input freqs has not the same number of frequencies as X")

    fig = plt.figure(figsize=(12, 7))
    fig.suptitle(title)
    for f in range(n_freqs):
        ax = plt.subplot((n_freqs - 1) // 8 + 1, 8, f + 1)
        plt.imshow(X[f], cmap=plt.get_cmap("Reds"))
        plt.title("{} Hz".format(freqs[f]))
        plt.xticks([])
        if ylabels and f == 0:
            plt.yticks(np.arange(0, len(ylabels), 2), ylabels[::2])
            ax.tick_params(axis="both", which="major", labelsize=7)
        elif ylabels and f == 8:
            plt.yticks(np.arange(1, len(ylabels), 2), ylabels[1::2])
            ax.tick_params(axis="both", which="major", labelsize=7)
        else:
            plt.yticks([])

    return fig


def plot_waveforms(X, display, *, times=None, color="gray", alpha=0.5,
                   linewidth=1.5, color_mean="k", color_std="gray", n_bins=50,
                   cmap=None):
    """Plot repetitions of a multichannel waveform.

    Parameters
    ----------
    X : ndarray, shape (n_reps, n_channels, n_times)
        Repetitions of the multichannel waveform.
    display : {"all", "mean", "mean+/-std", "hist"}
        Type of display:

        * "all" for all the repetitions;
        * "mean" for the mean of the repetitions;
        * "mean+/-std" for the mean +/- standard deviation of the repetitions;
        * "hist" for the 2D histogram of the repetitions.
    time : None | ndarray, shape (n_times,), default=None
        Values to display on x-axis.
    color : matplotlib color, optional
        Color of the lines, when ``display=all``.
    alpha : float, optional
        Alpha value used to cumulate repetitions, when ``display=all``.
    linewidth : float, optional
        Line width in points, when ``display=mean``.
    color_mean : matplotlib color, optional
        Color of the mean line, when ``display=mean``.
    color_std : matplotlib color, optional
        Color of the standard deviation area, when ``display=mean+/-std``.
    n_bins : int, optional
        Number of vertical bins for the 2D histogram, when ``display=hist``.
    cmap : Colormap or str, optional
        Color map for the histogram, when ``display=hist``.

    Returns
    -------
    fig : matplotlib figure
        Figure of waveform (one subplot by channel).

    Notes
    -----
    .. versionadded:: 0.3
    """
    if X.ndim != 3:
        raise ValueError("Input X has not 3 dimensions")
    n_reps, n_channels, n_times = X.shape
    if times is None:
        times = np.arange(n_times)
    elif times.shape != (n_times,):
        raise ValueError(
            "Parameter times has not the same number of values as X")

    fig, axes = plt.subplots(nrows=n_channels, ncols=1)
    if n_channels == 1:
        axes = [axes]
    channels = np.arange(n_channels)

    if display == "all":
        for (channel, ax) in zip(channels, axes):
            for i_rep in range(n_reps):
                ax.plot(times, X[i_rep, channel], c=color, alpha=alpha)

    elif display in ["mean", "mean+/-std"]:
        mean = np.mean(X, axis=0)
        for (channel, ax) in zip(channels, axes):
            ax.plot(times, mean[channel], c=color_mean, lw=linewidth)
        if display == "mean+/-std":
            std = np.std(X, axis=0)
            for (channel, ax) in zip(channels, axes):
                ax.fill_between(times, mean[channel] - std[channel],
                                mean[channel] + std[channel], color=color_std)

    elif display == "hist":
        times_rep = np.repeat(times[np.newaxis, :], n_reps, axis=0)
        for (channel, ax) in zip(channels, axes):
            ax.hist2d(times_rep.ravel(), X[:, channel, :].ravel(),
                      bins=(n_times, n_bins), cmap=cmap)

    else:
        raise ValueError(f"Unknown parameter display {display}")

    if n_channels > 1:
        for ax in axes[:-1]:
            ax.set_xticklabels([])  # remove xticklabels
    return fig


def _add_alpha(colors, alphas):
    """Add alphas to RGB channels"""
    try:
        from matplotlib.colors import to_rgb
    except ImportError:
        raise ImportError("Install matplotlib to add alpha")

    cols = [to_rgb(c) for c in colors]
    return [(c[0], c[1], c[2], a) for c, a in zip(cols, alphas[-len(cols):])]


def plot_cov_ellipse(ax, X, n_std=2.5, **kwds):
    """Plot 2x2 covariance matrix as an ellipse.

    Parameters
    ----------
    ax : matplotlib axis
        Axis of figure.
    X : ndarray, shape (2, 2)
        Covariance matrix.
    n_std : float, default=2.5
        Number of standard deviations.
    **kwds : dict
        Any further parameters are passed directly to the Ellipse.

    Returns
    -------
    ax : matplotlib axis
        Axis of figure.

    Notes
    -----
    .. versionadded:: 0.6

    References
    ----------
    .. [1] https://matplotlib.org/stable/gallery/statistics/confidence_ellipse.html
    """  # noqa
    if X.shape != (2, 2):
        raise ValueError("Input X must be a 2x2 covariance matrix")

    pearson = X[0, 1] / np.sqrt(X[0, 0] * X[1, 1])
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0), width=ell_radius_x * 2, height=ell_radius_y * 2,
                      facecolor='none', **kwds)
    scale_x = np.sqrt(X[0, 0]) * n_std
    scale_y = np.sqrt(X[1, 1]) * n_std
    transf = transforms.Affine2D().rotate_deg(45).scale(scale_x, scale_y)
    ellipse.set_transform(transf + ax.transData)

    return ax.add_patch(ellipse)


def plot_bihist(X, y, n_bins=10, title="Histogram"):
    """Plot histogram of bi-class predictions.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, 2)
        Predictions, distances or probabilities.
    y : ndarray, shape (n_matrices,)
        Labels for each matrix.
    n_bins : int, default=10
        Number of bins of histogram.
    title : str, default="Histogram"
        Title of figure.

    Returns
    -------
    fig : matplotlib figure
        Figure of histogram.

    Notes
    -----
    .. versionadded:: 0.6
    """
    if X.ndim != 2:
        raise ValueError("Input X has not 2 dimensions")
    if X.shape[1] != 2:
        raise ValueError("Input X has not 2 classes")

    classes = np.unique(y)
    if classes.shape[0] != 2:
        raise ValueError("Input y has not 2 labels")

    X = X / np.sum(X, axis=1, keepdims=True)
    X0 = X[y == classes[0], 0]
    X1 = 1 - X[y == classes[1], 1]

    def get_bins(X, n_bins, target=0.5):
        """Estimate bins with the garantee to have target value in bin edges"""
        bins = np.histogram_bin_edges(X, bins=n_bins)
        idx = (np.abs(bins - target)).argmin()
        bins[idx] = target
        return bins

    fig, ax = plt.subplots(figsize=(6, 5))
    ax.axvline(x=0.5, c="k", linestyle=":")
    ax.hist(X0, bins=get_bins(X0, n_bins), label=classes[0], alpha=0.5)
    ax.hist(X1, bins=get_bins(X1, n_bins), label=classes[1], alpha=0.5)

    (Xmin, Xmax) = ax.get_xlim()
    Xm = min(Xmin, 1 - Xmax)
    ax.set_xlim(Xm, 1 - Xm)
    ax.set(xlabel="Rescaled predictions", ylabel="Frequency", title=title)
    ax.legend(title="Classes", loc="upper left")

    return fig


def plot_biscatter(X, y):
    """Plot scatter of bi-class predictions.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, 2)
        Predictions, distances or probabilities.
    y : ndarray, shape (n_matrices,)
        Labels for each matrix.

    Returns
    -------
    fig : matplotlib figure
        Figure of scatter plot.

    Notes
    -----
    .. versionadded:: 0.6
    """

    if X.ndim != 2:
        raise ValueError("Input X has not 2 dimensions")
    if X.shape[1] != 2:
        raise ValueError("Input X has not 2 classes")

    classes = np.unique(y)
    if classes.shape[0] != 2:
        raise ValueError("Input y has not 2 labels")

    X0 = X[y == classes[0]]
    X1 = X[y == classes[1]]

    fig, ax = plt.subplots(figsize=(7, 7))
    ax.scatter(X0[:, 0], X0[:, 1], label=classes[0], alpha=1)
    ax.scatter(X1[:, 0], X1[:, 1], label=classes[1], alpha=0.5)
    ax.legend(title="Classes", loc="upper left")

    (Xmin, Xmax) = ax.get_xlim()
    (Ymin, Ymax) = ax.get_ylim()
    XYmin, XYmax = min(Xmin, Ymin), max(Xmax, Ymax)
    ax.plot([XYmin, XYmax], [XYmin, XYmax], c="k", linestyle=":")
    ax.set_xlim([XYmin, XYmax])
    ax.set_ylim([XYmin, XYmax])

    return fig
